sqlglot.dialects.clickhouse API documentation

Edit on GitHub

# sqlglot.dialects.clickhouse

View Source

      1from __future__ import annotations
      2
      3import typing as t
      4
      5from sqlglot import exp, generator, parser, tokens, transforms
      6from sqlglot.dialects.dialect import (
      7    Dialect,
      8    arg_max_or_min_no_count,
      9    inline_array_sql,
     10    no_pivot_sql,
     11    rename_func,
     12    var_map_sql,
     13)
     14from sqlglot.errors import ParseError
     15from sqlglot.helper import seq_get
     16from sqlglot.parser import parse_var_map
     17from sqlglot.tokens import Token, TokenType
     18
     19
     20def _lower_func(sql: str) -> str:
     21    index = sql.index("(")
     22    return sql[:index].lower() + sql[index:]
     23
     24
     25def _quantile_sql(self, e):
     26    quantile = e.args["quantile"]
     27    args = f"({self.sql(e, 'this')})"
     28    if isinstance(quantile, exp.Array):
     29        func = self.func("quantiles", *quantile)
     30    else:
     31        func = self.func("quantile", quantile)
     32    return func + args
     33
     34
     35class ClickHouse(Dialect):
     36    NORMALIZE_FUNCTIONS: bool | str = False
     37    NULL_ORDERING = "nulls_are_last"
     38    SUPPORTS_USER_DEFINED_TYPES = False
     39    SAFE_DIVISION = True
     40
     41    ESCAPE_SEQUENCES = {
     42        "\\0": "\0",
     43    }
     44
     45    class Tokenizer(tokens.Tokenizer):
     46        COMMENTS = ["--", "#", "#!", ("/*", "*/")]
     47        IDENTIFIERS = ['"', "`"]
     48        STRING_ESCAPES = ["'", "\\"]
     49        BIT_STRINGS = [("0b", "")]
     50        HEX_STRINGS = [("0x", ""), ("0X", "")]
     51        HEREDOC_STRINGS = ["$"]
     52
     53        KEYWORDS = {
     54            **tokens.Tokenizer.KEYWORDS,
     55            "ATTACH": TokenType.COMMAND,
     56            "DATETIME64": TokenType.DATETIME64,
     57            "DICTIONARY": TokenType.DICTIONARY,
     58            "ENUM": TokenType.ENUM,
     59            "ENUM8": TokenType.ENUM8,
     60            "ENUM16": TokenType.ENUM16,
     61            "FINAL": TokenType.FINAL,
     62            "FIXEDSTRING": TokenType.FIXEDSTRING,
     63            "FLOAT32": TokenType.FLOAT,
     64            "FLOAT64": TokenType.DOUBLE,
     65            "GLOBAL": TokenType.GLOBAL,
     66            "INT256": TokenType.INT256,
     67            "LOWCARDINALITY": TokenType.LOWCARDINALITY,
     68            "MAP": TokenType.MAP,
     69            "NESTED": TokenType.NESTED,
     70            "SAMPLE": TokenType.TABLE_SAMPLE,
     71            "TUPLE": TokenType.STRUCT,
     72            "UINT128": TokenType.UINT128,
     73            "UINT16": TokenType.USMALLINT,
     74            "UINT256": TokenType.UINT256,
     75            "UINT32": TokenType.UINT,
     76            "UINT64": TokenType.UBIGINT,
     77            "UINT8": TokenType.UTINYINT,
     78        }
     79
     80        SINGLE_TOKENS = {
     81            **tokens.Tokenizer.SINGLE_TOKENS,
     82            "$": TokenType.HEREDOC_STRING,
     83        }
     84
     85    class Parser(parser.Parser):
     86        # Tested in ClickHouse's playground, it seems that the following two queries do the same thing
     87        # * select x from t1 union all select x from t2 limit 1;
     88        # * select x from t1 union all (select x from t2 limit 1);
     89        MODIFIERS_ATTACHED_TO_UNION = False
     90
     91        FUNCTIONS = {
     92            **parser.Parser.FUNCTIONS,
     93            "ANY": exp.AnyValue.from_arg_list,
     94            "DATE_ADD": lambda args: exp.DateAdd(
     95                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
     96            ),
     97            "DATEADD": lambda args: exp.DateAdd(
     98                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
     99            ),
    100            "DATE_DIFF": lambda args: exp.DateDiff(
    101                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    102            ),
    103            "DATEDIFF": lambda args: exp.DateDiff(
    104                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    105            ),
    106            "MAP": parse_var_map,
    107            "MATCH": exp.RegexpLike.from_arg_list,
    108            "UNIQ": exp.ApproxDistinct.from_arg_list,
    109            "XOR": lambda args: exp.Xor(expressions=args),
    110        }
    111
    112        FUNCTIONS_WITH_ALIASED_ARGS = {*parser.Parser.FUNCTIONS_WITH_ALIASED_ARGS, "TUPLE"}
    113
    114        FUNCTION_PARSERS = {
    115            **parser.Parser.FUNCTION_PARSERS,
    116            "ARRAYJOIN": lambda self: self.expression(exp.Explode, this=self._parse_expression()),
    117            "QUANTILE": lambda self: self._parse_quantile(),
    118        }
    119
    120        FUNCTION_PARSERS.pop("MATCH")
    121
    122        NO_PAREN_FUNCTION_PARSERS = parser.Parser.NO_PAREN_FUNCTION_PARSERS.copy()
    123        NO_PAREN_FUNCTION_PARSERS.pop("ANY")
    124
    125        RANGE_PARSERS = {
    126            **parser.Parser.RANGE_PARSERS,
    127            TokenType.GLOBAL: lambda self, this: self._match(TokenType.IN)
    128            and self._parse_in(this, is_global=True),
    129        }
    130
    131        # The PLACEHOLDER entry is popped because 1) it doesn't affect Clickhouse (it corresponds to
    132        # the postgres-specific JSONBContains parser) and 2) it makes parsing the ternary op simpler.
    133        COLUMN_OPERATORS = parser.Parser.COLUMN_OPERATORS.copy()
    134        COLUMN_OPERATORS.pop(TokenType.PLACEHOLDER)
    135
    136        JOIN_KINDS = {
    137            *parser.Parser.JOIN_KINDS,
    138            TokenType.ANY,
    139            TokenType.ASOF,
    140            TokenType.ARRAY,
    141        }
    142
    143        TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS - {
    144            TokenType.ANY,
    145            TokenType.SETTINGS,
    146            TokenType.FORMAT,
    147            TokenType.ARRAY,
    148        }
    149
    150        LOG_DEFAULTS_TO_LN = True
    151
    152        QUERY_MODIFIER_PARSERS = {
    153            **parser.Parser.QUERY_MODIFIER_PARSERS,
    154            TokenType.SETTINGS: lambda self: (
    155                "settings",
    156                self._advance() or self._parse_csv(self._parse_conjunction),
    157            ),
    158            TokenType.FORMAT: lambda self: ("format", self._advance() or self._parse_id_var()),
    159        }
    160
    161        def _parse_conjunction(self) -> t.Optional[exp.Expression]:
    162            this = super()._parse_conjunction()
    163
    164            if self._match(TokenType.PLACEHOLDER):
    165                return self.expression(
    166                    exp.If,
    167                    this=this,
    168                    true=self._parse_conjunction(),
    169                    false=self._match(TokenType.COLON) and self._parse_conjunction(),
    170                )
    171
    172            return this
    173
    174        def _parse_placeholder(self) -> t.Optional[exp.Expression]:
    175            """
    176            Parse a placeholder expression like SELECT {abc: UInt32} or FROM {table: Identifier}
    177            https://clickhouse.com/docs/en/sql-reference/syntax#defining-and-using-query-parameters
    178            """
    179            if not self._match(TokenType.L_BRACE):
    180                return None
    181
    182            this = self._parse_id_var()
    183            self._match(TokenType.COLON)
    184            kind = self._parse_types(check_func=False, allow_identifiers=False) or (
    185                self._match_text_seq("IDENTIFIER") and "Identifier"
    186            )
    187
    188            if not kind:
    189                self.raise_error("Expecting a placeholder type or 'Identifier' for tables")
    190            elif not self._match(TokenType.R_BRACE):
    191                self.raise_error("Expecting }")
    192
    193            return self.expression(exp.Placeholder, this=this, kind=kind)
    194
    195        def _parse_in(self, this: t.Optional[exp.Expression], is_global: bool = False) -> exp.In:
    196            this = super()._parse_in(this)
    197            this.set("is_global", is_global)
    198            return this
    199
    200        def _parse_table(
    201            self,
    202            schema: bool = False,
    203            joins: bool = False,
    204            alias_tokens: t.Optional[t.Collection[TokenType]] = None,
    205            parse_bracket: bool = False,
    206        ) -> t.Optional[exp.Expression]:
    207            this = super()._parse_table(
    208                schema=schema, joins=joins, alias_tokens=alias_tokens, parse_bracket=parse_bracket
    209            )
    210
    211            if self._match(TokenType.FINAL):
    212                this = self.expression(exp.Final, this=this)
    213
    214            return this
    215
    216        def _parse_position(self, haystack_first: bool = False) -> exp.StrPosition:
    217            return super()._parse_position(haystack_first=True)
    218
    219        # https://clickhouse.com/docs/en/sql-reference/statements/select/with/
    220        def _parse_cte(self) -> exp.CTE:
    221            index = self._index
    222            try:
    223                # WITH <identifier> AS <subquery expression>
    224                return super()._parse_cte()
    225            except ParseError:
    226                # WITH <expression> AS <identifier>
    227                self._retreat(index)
    228
    229                return self.expression(
    230                    exp.CTE,
    231                    this=self._parse_field(),
    232                    alias=self._parse_table_alias(),
    233                    scalar=True,
    234                )
    235
    236        def _parse_join_parts(
    237            self,
    238        ) -> t.Tuple[t.Optional[Token], t.Optional[Token], t.Optional[Token]]:
    239            is_global = self._match(TokenType.GLOBAL) and self._prev
    240            kind_pre = self._match_set(self.JOIN_KINDS, advance=False) and self._prev
    241
    242            if kind_pre:
    243                kind = self._match_set(self.JOIN_KINDS) and self._prev
    244                side = self._match_set(self.JOIN_SIDES) and self._prev
    245                return is_global, side, kind
    246
    247            return (
    248                is_global,
    249                self._match_set(self.JOIN_SIDES) and self._prev,
    250                self._match_set(self.JOIN_KINDS) and self._prev,
    251            )
    252
    253        def _parse_join(
    254            self, skip_join_token: bool = False, parse_bracket: bool = False
    255        ) -> t.Optional[exp.Join]:
    256            join = super()._parse_join(skip_join_token=skip_join_token, parse_bracket=True)
    257
    258            if join:
    259                join.set("global", join.args.pop("method", None))
    260            return join
    261
    262        def _parse_function(
    263            self,
    264            functions: t.Optional[t.Dict[str, t.Callable]] = None,
    265            anonymous: bool = False,
    266            optional_parens: bool = True,
    267        ) -> t.Optional[exp.Expression]:
    268            func = super()._parse_function(
    269                functions=functions, anonymous=anonymous, optional_parens=optional_parens
    270            )
    271
    272            if isinstance(func, exp.Anonymous):
    273                params = self._parse_func_params(func)
    274
    275                if params:
    276                    return self.expression(
    277                        exp.ParameterizedAgg,
    278                        this=func.this,
    279                        expressions=func.expressions,
    280                        params=params,
    281                    )
    282
    283            return func
    284
    285        def _parse_func_params(
    286            self, this: t.Optional[exp.Func] = None
    287        ) -> t.Optional[t.List[exp.Expression]]:
    288            if self._match_pair(TokenType.R_PAREN, TokenType.L_PAREN):
    289                return self._parse_csv(self._parse_lambda)
    290
    291            if self._match(TokenType.L_PAREN):
    292                params = self._parse_csv(self._parse_lambda)
    293                self._match_r_paren(this)
    294                return params
    295
    296            return None
    297
    298        def _parse_quantile(self) -> exp.Quantile:
    299            this = self._parse_lambda()
    300            params = self._parse_func_params()
    301            if params:
    302                return self.expression(exp.Quantile, this=params[0], quantile=this)
    303            return self.expression(exp.Quantile, this=this, quantile=exp.Literal.number(0.5))
    304
    305        def _parse_wrapped_id_vars(self, optional: bool = False) -> t.List[exp.Expression]:
    306            return super()._parse_wrapped_id_vars(optional=True)
    307
    308        def _parse_primary_key(
    309            self, wrapped_optional: bool = False, in_props: bool = False
    310        ) -> exp.PrimaryKeyColumnConstraint | exp.PrimaryKey:
    311            return super()._parse_primary_key(
    312                wrapped_optional=wrapped_optional or in_props, in_props=in_props
    313            )
    314
    315        def _parse_on_property(self) -> t.Optional[exp.Expression]:
    316            index = self._index
    317            if self._match_text_seq("CLUSTER"):
    318                this = self._parse_id_var()
    319                if this:
    320                    return self.expression(exp.OnCluster, this=this)
    321                else:
    322                    self._retreat(index)
    323            return None
    324
    325    class Generator(generator.Generator):
    326        QUERY_HINTS = False
    327        STRUCT_DELIMITER = ("(", ")")
    328        NVL2_SUPPORTED = False
    329        TABLESAMPLE_REQUIRES_PARENS = False
    330
    331        STRING_TYPE_MAPPING = {
    332            exp.DataType.Type.CHAR: "String",
    333            exp.DataType.Type.LONGBLOB: "String",
    334            exp.DataType.Type.LONGTEXT: "String",
    335            exp.DataType.Type.MEDIUMBLOB: "String",
    336            exp.DataType.Type.MEDIUMTEXT: "String",
    337            exp.DataType.Type.TINYBLOB: "String",
    338            exp.DataType.Type.TINYTEXT: "String",
    339            exp.DataType.Type.TEXT: "String",
    340            exp.DataType.Type.VARBINARY: "String",
    341            exp.DataType.Type.VARCHAR: "String",
    342        }
    343
    344        TYPE_MAPPING = {
    345            **generator.Generator.TYPE_MAPPING,
    346            **STRING_TYPE_MAPPING,
    347            exp.DataType.Type.ARRAY: "Array",
    348            exp.DataType.Type.BIGINT: "Int64",
    349            exp.DataType.Type.DATETIME64: "DateTime64",
    350            exp.DataType.Type.DOUBLE: "Float64",
    351            exp.DataType.Type.ENUM: "Enum",
    352            exp.DataType.Type.ENUM8: "Enum8",
    353            exp.DataType.Type.ENUM16: "Enum16",
    354            exp.DataType.Type.FIXEDSTRING: "FixedString",
    355            exp.DataType.Type.FLOAT: "Float32",
    356            exp.DataType.Type.INT: "Int32",
    357            exp.DataType.Type.MEDIUMINT: "Int32",
    358            exp.DataType.Type.INT128: "Int128",
    359            exp.DataType.Type.INT256: "Int256",
    360            exp.DataType.Type.LOWCARDINALITY: "LowCardinality",
    361            exp.DataType.Type.MAP: "Map",
    362            exp.DataType.Type.NESTED: "Nested",
    363            exp.DataType.Type.NULLABLE: "Nullable",
    364            exp.DataType.Type.SMALLINT: "Int16",
    365            exp.DataType.Type.STRUCT: "Tuple",
    366            exp.DataType.Type.TINYINT: "Int8",
    367            exp.DataType.Type.UBIGINT: "UInt64",
    368            exp.DataType.Type.UINT: "UInt32",
    369            exp.DataType.Type.UINT128: "UInt128",
    370            exp.DataType.Type.UINT256: "UInt256",
    371            exp.DataType.Type.USMALLINT: "UInt16",
    372            exp.DataType.Type.UTINYINT: "UInt8",
    373        }
    374
    375        TRANSFORMS = {
    376            **generator.Generator.TRANSFORMS,
    377            exp.Select: transforms.preprocess([transforms.eliminate_qualify]),
    378            exp.AnyValue: rename_func("any"),
    379            exp.ApproxDistinct: rename_func("uniq"),
    380            exp.ArgMax: arg_max_or_min_no_count("argMax"),
    381            exp.ArgMin: arg_max_or_min_no_count("argMin"),
    382            exp.Array: inline_array_sql,
    383            exp.CastToStrType: rename_func("CAST"),
    384            exp.CurrentDate: lambda self, e: self.func("CURRENT_DATE"),
    385            exp.DateAdd: lambda self, e: self.func(
    386                "DATE_ADD", exp.Literal.string(e.text("unit") or "day"), e.expression, e.this
    387            ),
    388            exp.DateDiff: lambda self, e: self.func(
    389                "DATE_DIFF", exp.Literal.string(e.text("unit") or "day"), e.expression, e.this
    390            ),
    391            exp.Explode: rename_func("arrayJoin"),
    392            exp.Final: lambda self, e: f"{self.sql(e, 'this')} FINAL",
    393            exp.IsNan: rename_func("isNaN"),
    394            exp.Map: lambda self, e: _lower_func(var_map_sql(self, e)),
    395            exp.Nullif: rename_func("nullIf"),
    396            exp.PartitionedByProperty: lambda self, e: f"PARTITION BY {self.sql(e, 'this')}",
    397            exp.Pivot: no_pivot_sql,
    398            exp.Quantile: _quantile_sql,
    399            exp.RegexpLike: lambda self, e: f"match({self.format_args(e.this, e.expression)})",
    400            exp.StartsWith: rename_func("startsWith"),
    401            exp.StrPosition: lambda self, e: f"position({self.format_args(e.this, e.args.get('substr'), e.args.get('position'))})",
    402            exp.VarMap: lambda self, e: _lower_func(var_map_sql(self, e)),
    403            exp.Xor: lambda self, e: self.func("xor", e.this, e.expression, *e.expressions),
    404        }
    405
    406        PROPERTIES_LOCATION = {
    407            **generator.Generator.PROPERTIES_LOCATION,
    408            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    409            exp.PartitionedByProperty: exp.Properties.Location.POST_SCHEMA,
    410            exp.OnCluster: exp.Properties.Location.POST_NAME,
    411        }
    412
    413        JOIN_HINTS = False
    414        TABLE_HINTS = False
    415        EXPLICIT_UNION = True
    416        GROUPINGS_SEP = ""
    417
    418        # there's no list in docs, but it can be found in Clickhouse code
    419        # see `ClickHouse/src/Parsers/ParserCreate*.cpp`
    420        ON_CLUSTER_TARGETS = {
    421            "DATABASE",
    422            "TABLE",
    423            "VIEW",
    424            "DICTIONARY",
    425            "INDEX",
    426            "FUNCTION",
    427            "NAMED COLLECTION",
    428        }
    429
    430        def _any_to_has(
    431            self,
    432            expression: exp.EQ | exp.NEQ,
    433            default: t.Callable[[t.Any], str],
    434            prefix: str = "",
    435        ) -> str:
    436            if isinstance(expression.left, exp.Any):
    437                arr = expression.left
    438                this = expression.right
    439            elif isinstance(expression.right, exp.Any):
    440                arr = expression.right
    441                this = expression.left
    442            else:
    443                return default(expression)
    444            return prefix + self.func("has", arr.this.unnest(), this)
    445
    446        def eq_sql(self, expression: exp.EQ) -> str:
    447            return self._any_to_has(expression, super().eq_sql)
    448
    449        def neq_sql(self, expression: exp.NEQ) -> str:
    450            return self._any_to_has(expression, super().neq_sql, "NOT ")
    451
    452        def regexpilike_sql(self, expression: exp.RegexpILike) -> str:
    453            # Manually add a flag to make the search case-insensitive
    454            regex = self.func("CONCAT", "'(?i)'", expression.expression)
    455            return f"match({self.format_args(expression.this, regex)})"
    456
    457        def datatype_sql(self, expression: exp.DataType) -> str:
    458            # String is the standard ClickHouse type, every other variant is just an alias.
    459            # Additionally, any supplied length parameter will be ignored.
    460            #
    461            # https://clickhouse.com/docs/en/sql-reference/data-types/string
    462            if expression.this in self.STRING_TYPE_MAPPING:
    463                return "String"
    464
    465            return super().datatype_sql(expression)
    466
    467        def cte_sql(self, expression: exp.CTE) -> str:
    468            if expression.args.get("scalar"):
    469                this = self.sql(expression, "this")
    470                alias = self.sql(expression, "alias")
    471                return f"{this} AS {alias}"
    472
    473            return super().cte_sql(expression)
    474
    475        def after_limit_modifiers(self, expression: exp.Expression) -> t.List[str]:
    476            return super().after_limit_modifiers(expression) + [
    477                self.seg("SETTINGS ") + self.expressions(expression, key="settings", flat=True)
    478                if expression.args.get("settings")
    479                else "",
    480                self.seg("FORMAT ") + self.sql(expression, "format")
    481                if expression.args.get("format")
    482                else "",
    483            ]
    484
    485        def parameterizedagg_sql(self, expression: exp.Anonymous) -> str:
    486            params = self.expressions(expression, key="params", flat=True)
    487            return self.func(expression.name, *expression.expressions) + f"({params})"
    488
    489        def placeholder_sql(self, expression: exp.Placeholder) -> str:
    490            return f"{{{expression.name}: {self.sql(expression, 'kind')}}}"
    491
    492        def oncluster_sql(self, expression: exp.OnCluster) -> str:
    493            return f"ON CLUSTER {self.sql(expression, 'this')}"
    494
    495        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
    496            kind = self.sql(expression, "kind").upper()
    497            if kind in self.ON_CLUSTER_TARGETS and locations.get(exp.Properties.Location.POST_NAME):
    498                this_name = self.sql(expression.this, "this")
    499                this_properties = " ".join(
    500                    [self.sql(prop) for prop in locations[exp.Properties.Location.POST_NAME]]
    501                )
    502                this_schema = self.schema_columns_sql(expression.this)
    503                return f"{this_name}{self.sep()}{this_properties}{self.sep()}{this_schema}"
    504
    505            return super().createable_sql(expression, locations)

class ClickHouse(sqlglot.dialects.dialect.Dialect): View Source

     36class ClickHouse(Dialect):
     37    NORMALIZE_FUNCTIONS: bool | str = False
     38    NULL_ORDERING = "nulls_are_last"
     39    SUPPORTS_USER_DEFINED_TYPES = False
     40    SAFE_DIVISION = True
     41
     42    ESCAPE_SEQUENCES = {
     43        "\\0": "\0",
     44    }
     45
     46    class Tokenizer(tokens.Tokenizer):
     47        COMMENTS = ["--", "#", "#!", ("/*", "*/")]
     48        IDENTIFIERS = ['"', "`"]
     49        STRING_ESCAPES = ["'", "\\"]
     50        BIT_STRINGS = [("0b", "")]
     51        HEX_STRINGS = [("0x", ""), ("0X", "")]
     52        HEREDOC_STRINGS = ["$"]
     53
     54        KEYWORDS = {
     55            **tokens.Tokenizer.KEYWORDS,
     56            "ATTACH": TokenType.COMMAND,
     57            "DATETIME64": TokenType.DATETIME64,
     58            "DICTIONARY": TokenType.DICTIONARY,
     59            "ENUM": TokenType.ENUM,
     60            "ENUM8": TokenType.ENUM8,
     61            "ENUM16": TokenType.ENUM16,
     62            "FINAL": TokenType.FINAL,
     63            "FIXEDSTRING": TokenType.FIXEDSTRING,
     64            "FLOAT32": TokenType.FLOAT,
     65            "FLOAT64": TokenType.DOUBLE,
     66            "GLOBAL": TokenType.GLOBAL,
     67            "INT256": TokenType.INT256,
     68            "LOWCARDINALITY": TokenType.LOWCARDINALITY,
     69            "MAP": TokenType.MAP,
     70            "NESTED": TokenType.NESTED,
     71            "SAMPLE": TokenType.TABLE_SAMPLE,
     72            "TUPLE": TokenType.STRUCT,
     73            "UINT128": TokenType.UINT128,
     74            "UINT16": TokenType.USMALLINT,
     75            "UINT256": TokenType.UINT256,
     76            "UINT32": TokenType.UINT,
     77            "UINT64": TokenType.UBIGINT,
     78            "UINT8": TokenType.UTINYINT,
     79        }
     80
     81        SINGLE_TOKENS = {
     82            **tokens.Tokenizer.SINGLE_TOKENS,
     83            "$": TokenType.HEREDOC_STRING,
     84        }
     85
     86    class Parser(parser.Parser):
     87        # Tested in ClickHouse's playground, it seems that the following two queries do the same thing
     88        # * select x from t1 union all select x from t2 limit 1;
     89        # * select x from t1 union all (select x from t2 limit 1);
     90        MODIFIERS_ATTACHED_TO_UNION = False
     91
     92        FUNCTIONS = {
     93            **parser.Parser.FUNCTIONS,
     94            "ANY": exp.AnyValue.from_arg_list,
     95            "DATE_ADD": lambda args: exp.DateAdd(
     96                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
     97            ),
     98            "DATEADD": lambda args: exp.DateAdd(
     99                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    100            ),
    101            "DATE_DIFF": lambda args: exp.DateDiff(
    102                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    103            ),
    104            "DATEDIFF": lambda args: exp.DateDiff(
    105                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    106            ),
    107            "MAP": parse_var_map,
    108            "MATCH": exp.RegexpLike.from_arg_list,
    109            "UNIQ": exp.ApproxDistinct.from_arg_list,
    110            "XOR": lambda args: exp.Xor(expressions=args),
    111        }
    112
    113        FUNCTIONS_WITH_ALIASED_ARGS = {*parser.Parser.FUNCTIONS_WITH_ALIASED_ARGS, "TUPLE"}
    114
    115        FUNCTION_PARSERS = {
    116            **parser.Parser.FUNCTION_PARSERS,
    117            "ARRAYJOIN": lambda self: self.expression(exp.Explode, this=self._parse_expression()),
    118            "QUANTILE": lambda self: self._parse_quantile(),
    119        }
    120
    121        FUNCTION_PARSERS.pop("MATCH")
    122
    123        NO_PAREN_FUNCTION_PARSERS = parser.Parser.NO_PAREN_FUNCTION_PARSERS.copy()
    124        NO_PAREN_FUNCTION_PARSERS.pop("ANY")
    125
    126        RANGE_PARSERS = {
    127            **parser.Parser.RANGE_PARSERS,
    128            TokenType.GLOBAL: lambda self, this: self._match(TokenType.IN)
    129            and self._parse_in(this, is_global=True),
    130        }
    131
    132        # The PLACEHOLDER entry is popped because 1) it doesn't affect Clickhouse (it corresponds to
    133        # the postgres-specific JSONBContains parser) and 2) it makes parsing the ternary op simpler.
    134        COLUMN_OPERATORS = parser.Parser.COLUMN_OPERATORS.copy()
    135        COLUMN_OPERATORS.pop(TokenType.PLACEHOLDER)
    136
    137        JOIN_KINDS = {
    138            *parser.Parser.JOIN_KINDS,
    139            TokenType.ANY,
    140            TokenType.ASOF,
    141            TokenType.ARRAY,
    142        }
    143
    144        TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS - {
    145            TokenType.ANY,
    146            TokenType.SETTINGS,
    147            TokenType.FORMAT,
    148            TokenType.ARRAY,
    149        }
    150
    151        LOG_DEFAULTS_TO_LN = True
    152
    153        QUERY_MODIFIER_PARSERS = {
    154            **parser.Parser.QUERY_MODIFIER_PARSERS,
    155            TokenType.SETTINGS: lambda self: (
    156                "settings",
    157                self._advance() or self._parse_csv(self._parse_conjunction),
    158            ),
    159            TokenType.FORMAT: lambda self: ("format", self._advance() or self._parse_id_var()),
    160        }
    161
    162        def _parse_conjunction(self) -> t.Optional[exp.Expression]:
    163            this = super()._parse_conjunction()
    164
    165            if self._match(TokenType.PLACEHOLDER):
    166                return self.expression(
    167                    exp.If,
    168                    this=this,
    169                    true=self._parse_conjunction(),
    170                    false=self._match(TokenType.COLON) and self._parse_conjunction(),
    171                )
    172
    173            return this
    174
    175        def _parse_placeholder(self) -> t.Optional[exp.Expression]:
    176            """
    177            Parse a placeholder expression like SELECT {abc: UInt32} or FROM {table: Identifier}
    178            https://clickhouse.com/docs/en/sql-reference/syntax#defining-and-using-query-parameters
    179            """
    180            if not self._match(TokenType.L_BRACE):
    181                return None
    182
    183            this = self._parse_id_var()
    184            self._match(TokenType.COLON)
    185            kind = self._parse_types(check_func=False, allow_identifiers=False) or (
    186                self._match_text_seq("IDENTIFIER") and "Identifier"
    187            )
    188
    189            if not kind:
    190                self.raise_error("Expecting a placeholder type or 'Identifier' for tables")
    191            elif not self._match(TokenType.R_BRACE):
    192                self.raise_error("Expecting }")
    193
    194            return self.expression(exp.Placeholder, this=this, kind=kind)
    195
    196        def _parse_in(self, this: t.Optional[exp.Expression], is_global: bool = False) -> exp.In:
    197            this = super()._parse_in(this)
    198            this.set("is_global", is_global)
    199            return this
    200
    201        def _parse_table(
    202            self,
    203            schema: bool = False,
    204            joins: bool = False,
    205            alias_tokens: t.Optional[t.Collection[TokenType]] = None,
    206            parse_bracket: bool = False,
    207        ) -> t.Optional[exp.Expression]:
    208            this = super()._parse_table(
    209                schema=schema, joins=joins, alias_tokens=alias_tokens, parse_bracket=parse_bracket
    210            )
    211
    212            if self._match(TokenType.FINAL):
    213                this = self.expression(exp.Final, this=this)
    214
    215            return this
    216
    217        def _parse_position(self, haystack_first: bool = False) -> exp.StrPosition:
    218            return super()._parse_position(haystack_first=True)
    219
    220        # https://clickhouse.com/docs/en/sql-reference/statements/select/with/
    221        def _parse_cte(self) -> exp.CTE:
    222            index = self._index
    223            try:
    224                # WITH <identifier> AS <subquery expression>
    225                return super()._parse_cte()
    226            except ParseError:
    227                # WITH <expression> AS <identifier>
    228                self._retreat(index)
    229
    230                return self.expression(
    231                    exp.CTE,
    232                    this=self._parse_field(),
    233                    alias=self._parse_table_alias(),
    234                    scalar=True,
    235                )
    236
    237        def _parse_join_parts(
    238            self,
    239        ) -> t.Tuple[t.Optional[Token], t.Optional[Token], t.Optional[Token]]:
    240            is_global = self._match(TokenType.GLOBAL) and self._prev
    241            kind_pre = self._match_set(self.JOIN_KINDS, advance=False) and self._prev
    242
    243            if kind_pre:
    244                kind = self._match_set(self.JOIN_KINDS) and self._prev
    245                side = self._match_set(self.JOIN_SIDES) and self._prev
    246                return is_global, side, kind
    247
    248            return (
    249                is_global,
    250                self._match_set(self.JOIN_SIDES) and self._prev,
    251                self._match_set(self.JOIN_KINDS) and self._prev,
    252            )
    253
    254        def _parse_join(
    255            self, skip_join_token: bool = False, parse_bracket: bool = False
    256        ) -> t.Optional[exp.Join]:
    257            join = super()._parse_join(skip_join_token=skip_join_token, parse_bracket=True)
    258
    259            if join:
    260                join.set("global", join.args.pop("method", None))
    261            return join
    262
    263        def _parse_function(
    264            self,
    265            functions: t.Optional[t.Dict[str, t.Callable]] = None,
    266            anonymous: bool = False,
    267            optional_parens: bool = True,
    268        ) -> t.Optional[exp.Expression]:
    269            func = super()._parse_function(
    270                functions=functions, anonymous=anonymous, optional_parens=optional_parens
    271            )
    272
    273            if isinstance(func, exp.Anonymous):
    274                params = self._parse_func_params(func)
    275
    276                if params:
    277                    return self.expression(
    278                        exp.ParameterizedAgg,
    279                        this=func.this,
    280                        expressions=func.expressions,
    281                        params=params,
    282                    )
    283
    284            return func
    285
    286        def _parse_func_params(
    287            self, this: t.Optional[exp.Func] = None
    288        ) -> t.Optional[t.List[exp.Expression]]:
    289            if self._match_pair(TokenType.R_PAREN, TokenType.L_PAREN):
    290                return self._parse_csv(self._parse_lambda)
    291
    292            if self._match(TokenType.L_PAREN):
    293                params = self._parse_csv(self._parse_lambda)
    294                self._match_r_paren(this)
    295                return params
    296
    297            return None
    298
    299        def _parse_quantile(self) -> exp.Quantile:
    300            this = self._parse_lambda()
    301            params = self._parse_func_params()
    302            if params:
    303                return self.expression(exp.Quantile, this=params[0], quantile=this)
    304            return self.expression(exp.Quantile, this=this, quantile=exp.Literal.number(0.5))
    305
    306        def _parse_wrapped_id_vars(self, optional: bool = False) -> t.List[exp.Expression]:
    307            return super()._parse_wrapped_id_vars(optional=True)
    308
    309        def _parse_primary_key(
    310            self, wrapped_optional: bool = False, in_props: bool = False
    311        ) -> exp.PrimaryKeyColumnConstraint | exp.PrimaryKey:
    312            return super()._parse_primary_key(
    313                wrapped_optional=wrapped_optional or in_props, in_props=in_props
    314            )
    315
    316        def _parse_on_property(self) -> t.Optional[exp.Expression]:
    317            index = self._index
    318            if self._match_text_seq("CLUSTER"):
    319                this = self._parse_id_var()
    320                if this:
    321                    return self.expression(exp.OnCluster, this=this)
    322                else:
    323                    self._retreat(index)
    324            return None
    325
    326    class Generator(generator.Generator):
    327        QUERY_HINTS = False
    328        STRUCT_DELIMITER = ("(", ")")
    329        NVL2_SUPPORTED = False
    330        TABLESAMPLE_REQUIRES_PARENS = False
    331
    332        STRING_TYPE_MAPPING = {
    333            exp.DataType.Type.CHAR: "String",
    334            exp.DataType.Type.LONGBLOB: "String",
    335            exp.DataType.Type.LONGTEXT: "String",
    336            exp.DataType.Type.MEDIUMBLOB: "String",
    337            exp.DataType.Type.MEDIUMTEXT: "String",
    338            exp.DataType.Type.TINYBLOB: "String",
    339            exp.DataType.Type.TINYTEXT: "String",
    340            exp.DataType.Type.TEXT: "String",
    341            exp.DataType.Type.VARBINARY: "String",
    342            exp.DataType.Type.VARCHAR: "String",
    343        }
    344
    345        TYPE_MAPPING = {
    346            **generator.Generator.TYPE_MAPPING,
    347            **STRING_TYPE_MAPPING,
    348            exp.DataType.Type.ARRAY: "Array",
    349            exp.DataType.Type.BIGINT: "Int64",
    350            exp.DataType.Type.DATETIME64: "DateTime64",
    351            exp.DataType.Type.DOUBLE: "Float64",
    352            exp.DataType.Type.ENUM: "Enum",
    353            exp.DataType.Type.ENUM8: "Enum8",
    354            exp.DataType.Type.ENUM16: "Enum16",
    355            exp.DataType.Type.FIXEDSTRING: "FixedString",
    356            exp.DataType.Type.FLOAT: "Float32",
    357            exp.DataType.Type.INT: "Int32",
    358            exp.DataType.Type.MEDIUMINT: "Int32",
    359            exp.DataType.Type.INT128: "Int128",
    360            exp.DataType.Type.INT256: "Int256",
    361            exp.DataType.Type.LOWCARDINALITY: "LowCardinality",
    362            exp.DataType.Type.MAP: "Map",
    363            exp.DataType.Type.NESTED: "Nested",
    364            exp.DataType.Type.NULLABLE: "Nullable",
    365            exp.DataType.Type.SMALLINT: "Int16",
    366            exp.DataType.Type.STRUCT: "Tuple",
    367            exp.DataType.Type.TINYINT: "Int8",
    368            exp.DataType.Type.UBIGINT: "UInt64",
    369            exp.DataType.Type.UINT: "UInt32",
    370            exp.DataType.Type.UINT128: "UInt128",
    371            exp.DataType.Type.UINT256: "UInt256",
    372            exp.DataType.Type.USMALLINT: "UInt16",
    373            exp.DataType.Type.UTINYINT: "UInt8",
    374        }
    375
    376        TRANSFORMS = {
    377            **generator.Generator.TRANSFORMS,
    378            exp.Select: transforms.preprocess([transforms.eliminate_qualify]),
    379            exp.AnyValue: rename_func("any"),
    380            exp.ApproxDistinct: rename_func("uniq"),
    381            exp.ArgMax: arg_max_or_min_no_count("argMax"),
    382            exp.ArgMin: arg_max_or_min_no_count("argMin"),
    383            exp.Array: inline_array_sql,
    384            exp.CastToStrType: rename_func("CAST"),
    385            exp.CurrentDate: lambda self, e: self.func("CURRENT_DATE"),
    386            exp.DateAdd: lambda self, e: self.func(
    387                "DATE_ADD", exp.Literal.string(e.text("unit") or "day"), e.expression, e.this
    388            ),
    389            exp.DateDiff: lambda self, e: self.func(
    390                "DATE_DIFF", exp.Literal.string(e.text("unit") or "day"), e.expression, e.this
    391            ),
    392            exp.Explode: rename_func("arrayJoin"),
    393            exp.Final: lambda self, e: f"{self.sql(e, 'this')} FINAL",
    394            exp.IsNan: rename_func("isNaN"),
    395            exp.Map: lambda self, e: _lower_func(var_map_sql(self, e)),
    396            exp.Nullif: rename_func("nullIf"),
    397            exp.PartitionedByProperty: lambda self, e: f"PARTITION BY {self.sql(e, 'this')}",
    398            exp.Pivot: no_pivot_sql,
    399            exp.Quantile: _quantile_sql,
    400            exp.RegexpLike: lambda self, e: f"match({self.format_args(e.this, e.expression)})",
    401            exp.StartsWith: rename_func("startsWith"),
    402            exp.StrPosition: lambda self, e: f"position({self.format_args(e.this, e.args.get('substr'), e.args.get('position'))})",
    403            exp.VarMap: lambda self, e: _lower_func(var_map_sql(self, e)),
    404            exp.Xor: lambda self, e: self.func("xor", e.this, e.expression, *e.expressions),
    405        }
    406
    407        PROPERTIES_LOCATION = {
    408            **generator.Generator.PROPERTIES_LOCATION,
    409            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    410            exp.PartitionedByProperty: exp.Properties.Location.POST_SCHEMA,
    411            exp.OnCluster: exp.Properties.Location.POST_NAME,
    412        }
    413
    414        JOIN_HINTS = False
    415        TABLE_HINTS = False
    416        EXPLICIT_UNION = True
    417        GROUPINGS_SEP = ""
    418
    419        # there's no list in docs, but it can be found in Clickhouse code
    420        # see `ClickHouse/src/Parsers/ParserCreate*.cpp`
    421        ON_CLUSTER_TARGETS = {
    422            "DATABASE",
    423            "TABLE",
    424            "VIEW",
    425            "DICTIONARY",
    426            "INDEX",
    427            "FUNCTION",
    428            "NAMED COLLECTION",
    429        }
    430
    431        def _any_to_has(
    432            self,
    433            expression: exp.EQ | exp.NEQ,
    434            default: t.Callable[[t.Any], str],
    435            prefix: str = "",
    436        ) -> str:
    437            if isinstance(expression.left, exp.Any):
    438                arr = expression.left
    439                this = expression.right
    440            elif isinstance(expression.right, exp.Any):
    441                arr = expression.right
    442                this = expression.left
    443            else:
    444                return default(expression)
    445            return prefix + self.func("has", arr.this.unnest(), this)
    446
    447        def eq_sql(self, expression: exp.EQ) -> str:
    448            return self._any_to_has(expression, super().eq_sql)
    449
    450        def neq_sql(self, expression: exp.NEQ) -> str:
    451            return self._any_to_has(expression, super().neq_sql, "NOT ")
    452
    453        def regexpilike_sql(self, expression: exp.RegexpILike) -> str:
    454            # Manually add a flag to make the search case-insensitive
    455            regex = self.func("CONCAT", "'(?i)'", expression.expression)
    456            return f"match({self.format_args(expression.this, regex)})"
    457
    458        def datatype_sql(self, expression: exp.DataType) -> str:
    459            # String is the standard ClickHouse type, every other variant is just an alias.
    460            # Additionally, any supplied length parameter will be ignored.
    461            #
    462            # https://clickhouse.com/docs/en/sql-reference/data-types/string
    463            if expression.this in self.STRING_TYPE_MAPPING:
    464                return "String"
    465
    466            return super().datatype_sql(expression)
    467
    468        def cte_sql(self, expression: exp.CTE) -> str:
    469            if expression.args.get("scalar"):
    470                this = self.sql(expression, "this")
    471                alias = self.sql(expression, "alias")
    472                return f"{this} AS {alias}"
    473
    474            return super().cte_sql(expression)
    475
    476        def after_limit_modifiers(self, expression: exp.Expression) -> t.List[str]:
    477            return super().after_limit_modifiers(expression) + [
    478                self.seg("SETTINGS ") + self.expressions(expression, key="settings", flat=True)
    479                if expression.args.get("settings")
    480                else "",
    481                self.seg("FORMAT ") + self.sql(expression, "format")
    482                if expression.args.get("format")
    483                else "",
    484            ]
    485
    486        def parameterizedagg_sql(self, expression: exp.Anonymous) -> str:
    487            params = self.expressions(expression, key="params", flat=True)
    488            return self.func(expression.name, *expression.expressions) + f"({params})"
    489
    490        def placeholder_sql(self, expression: exp.Placeholder) -> str:
    491            return f"{{{expression.name}: {self.sql(expression, 'kind')}}}"
    492
    493        def oncluster_sql(self, expression: exp.OnCluster) -> str:
    494            return f"ON CLUSTER {self.sql(expression, 'this')}"
    495
    496        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
    497            kind = self.sql(expression, "kind").upper()
    498            if kind in self.ON_CLUSTER_TARGETS and locations.get(exp.Properties.Location.POST_NAME):
    499                this_name = self.sql(expression.this, "this")
    500                this_properties = " ".join(
    501                    [self.sql(prop) for prop in locations[exp.Properties.Location.POST_NAME]]
    502                )
    503                this_schema = self.schema_columns_sql(expression.this)
    504                return f"{this_name}{self.sep()}{this_properties}{self.sep()}{this_schema}"
    505
    506            return super().createable_sql(expression, locations)

NORMALIZE_FUNCTIONS: bool | str = False
NULL_ORDERING = 'nulls_are_last'
SUPPORTS_USER_DEFINED_TYPES = False
SAFE_DIVISION = True
ESCAPE_SEQUENCES = {'\\0': '\x00'}
tokenizer_class = <class 'ClickHouse.Tokenizer'>
parser_class = <class 'ClickHouse.Parser'>
generator_class = <class 'ClickHouse.Generator'>
TIME_TRIE: Dict = {}
FORMAT_TRIE: Dict = {}
INVERSE_TIME_MAPPING: Dict[str, str] = {}
INVERSE_TIME_TRIE: Dict = {}
INVERSE_ESCAPE_SEQUENCES: Dict[str, str] = {'\x00': '\\0'}
QUOTE_START = "'"
QUOTE_END = "'"
IDENTIFIER_START = '"'
IDENTIFIER_END = '"'
BIT_START: Optional[str] = '0b'
BIT_END: Optional[str] = ''
HEX_START: Optional[str] = '0x'
HEX_END: Optional[str] = ''
BYTE_START: Optional[str] = None
BYTE_END: Optional[str] = None

##### Inherited Members

sqlglot.dialects.dialect.Dialect
  Dialect
  INDEX_OFFSET
  UNNEST_COLUMN_ONLY
  ALIAS_POST_TABLESAMPLE
  NORMALIZATION_STRATEGY
  IDENTIFIERS_CAN_START_WITH_DIGIT
  DPIPE_IS_STRING_CONCAT
  STRICT_STRING_CONCAT
  SUPPORTS_SEMI_ANTI_JOIN
  LOG_BASE_FIRST
  TYPED_DIVISION
  CONCAT_COALESCE
  DATE_FORMAT
  DATEINT_FORMAT
  TIME_FORMAT
  TIME_MAPPING
  FORMAT_MAPPING
  PSEUDOCOLUMNS
  get_or_raise
  format_time
  normalize_identifier
  case_sensitive
  can_identify
  quote_identifier
  parse
  parse_into
  generate
  transpile
  tokenize
  tokenizer
  parser
  generator

class ClickHouse.Tokenizer(sqlglot.tokens.Tokenizer): View Source

    46    class Tokenizer(tokens.Tokenizer):
    47        COMMENTS = ["--", "#", "#!", ("/*", "*/")]
    48        IDENTIFIERS = ['"', "`"]
    49        STRING_ESCAPES = ["'", "\\"]
    50        BIT_STRINGS = [("0b", "")]
    51        HEX_STRINGS = [("0x", ""), ("0X", "")]
    52        HEREDOC_STRINGS = ["$"]
    53
    54        KEYWORDS = {
    55            **tokens.Tokenizer.KEYWORDS,
    56            "ATTACH": TokenType.COMMAND,
    57            "DATETIME64": TokenType.DATETIME64,
    58            "DICTIONARY": TokenType.DICTIONARY,
    59            "ENUM": TokenType.ENUM,
    60            "ENUM8": TokenType.ENUM8,
    61            "ENUM16": TokenType.ENUM16,
    62            "FINAL": TokenType.FINAL,
    63            "FIXEDSTRING": TokenType.FIXEDSTRING,
    64            "FLOAT32": TokenType.FLOAT,
    65            "FLOAT64": TokenType.DOUBLE,
    66            "GLOBAL": TokenType.GLOBAL,
    67            "INT256": TokenType.INT256,
    68            "LOWCARDINALITY": TokenType.LOWCARDINALITY,
    69            "MAP": TokenType.MAP,
    70            "NESTED": TokenType.NESTED,
    71            "SAMPLE": TokenType.TABLE_SAMPLE,
    72            "TUPLE": TokenType.STRUCT,
    73            "UINT128": TokenType.UINT128,
    74            "UINT16": TokenType.USMALLINT,
    75            "UINT256": TokenType.UINT256,
    76            "UINT32": TokenType.UINT,
    77            "UINT64": TokenType.UBIGINT,
    78            "UINT8": TokenType.UTINYINT,
    79        }
    80
    81        SINGLE_TOKENS = {
    82            **tokens.Tokenizer.SINGLE_TOKENS,
    83            "$": TokenType.HEREDOC_STRING,
    84        }

COMMENTS = ['--', '#', '#!', ('/*', '*/')]
IDENTIFIERS = ['"', '`']
STRING_ESCAPES = ["'", '\\']
BIT_STRINGS = [('0b', '')]
HEX_STRINGS = [('0x', ''), ('0X', '')]
HEREDOC_STRINGS = ['$']
KEYWORDS = {'{%': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%-': <TokenType.BLOCK_START: 'BLOCK_START'>, '%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '+%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '{{+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{{-': <TokenType.BLOCK_START: 'BLOCK_START'>, '+}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '/*+': <TokenType.HINT: 'HINT'>, '==': <TokenType.EQ: 'EQ'>, '::': <TokenType.DCOLON: 'DCOLON'>, '||': <TokenType.DPIPE: 'DPIPE'>, '>=': <TokenType.GTE: 'GTE'>, '<=': <TokenType.LTE: 'LTE'>, '<>': <TokenType.NEQ: 'NEQ'>, '!=': <TokenType.NEQ: 'NEQ'>, ':=': <TokenType.COLON_EQ: 'COLON_EQ'>, '<=>': <TokenType.NULLSAFE_EQ: 'NULLSAFE_EQ'>, '->': <TokenType.ARROW: 'ARROW'>, '->>': <TokenType.DARROW: 'DARROW'>, '=>': <TokenType.FARROW: 'FARROW'>, '#>': <TokenType.HASH_ARROW: 'HASH_ARROW'>, '#>>': <TokenType.DHASH_ARROW: 'DHASH_ARROW'>, '<->': <TokenType.LR_ARROW: 'LR_ARROW'>, '&&': <TokenType.DAMP: 'DAMP'>, '??': <TokenType.DQMARK: 'DQMARK'>, 'ALL': <TokenType.ALL: 'ALL'>, 'ALWAYS': <TokenType.ALWAYS: 'ALWAYS'>, 'AND': <TokenType.AND: 'AND'>, 'ANTI': <TokenType.ANTI: 'ANTI'>, 'ANY': <TokenType.ANY: 'ANY'>, 'ASC': <TokenType.ASC: 'ASC'>, 'AS': <TokenType.ALIAS: 'ALIAS'>, 'ASOF': <TokenType.ASOF: 'ASOF'>, 'AUTOINCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'AUTO_INCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'BEGIN': <TokenType.BEGIN: 'BEGIN'>, 'BETWEEN': <TokenType.BETWEEN: 'BETWEEN'>, 'CACHE': <TokenType.CACHE: 'CACHE'>, 'UNCACHE': <TokenType.UNCACHE: 'UNCACHE'>, 'CASE': <TokenType.CASE: 'CASE'>, 'CHARACTER SET': <TokenType.CHARACTER_SET: 'CHARACTER_SET'>, 'CLUSTER BY': <TokenType.CLUSTER_BY: 'CLUSTER_BY'>, 'COLLATE': <TokenType.COLLATE: 'COLLATE'>, 'COLUMN': <TokenType.COLUMN: 'COLUMN'>, 'COMMIT': <TokenType.COMMIT: 'COMMIT'>, 'CONNECT BY': <TokenType.CONNECT_BY: 'CONNECT_BY'>, 'CONSTRAINT': <TokenType.CONSTRAINT: 'CONSTRAINT'>, 'CREATE': <TokenType.CREATE: 'CREATE'>, 'CROSS': <TokenType.CROSS: 'CROSS'>, 'CUBE': <TokenType.CUBE: 'CUBE'>, 'CURRENT_DATE': <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, 'CURRENT_TIME': <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, 'CURRENT_TIMESTAMP': <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, 'CURRENT_USER': <TokenType.CURRENT_USER: 'CURRENT_USER'>, 'DATABASE': <TokenType.DATABASE: 'DATABASE'>, 'DEFAULT': <TokenType.DEFAULT: 'DEFAULT'>, 'DELETE': <TokenType.DELETE: 'DELETE'>, 'DESC': <TokenType.DESC: 'DESC'>, 'DESCRIBE': <TokenType.DESCRIBE: 'DESCRIBE'>, 'DISTINCT': <TokenType.DISTINCT: 'DISTINCT'>, 'DISTRIBUTE BY': <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>, 'DIV': <TokenType.DIV: 'DIV'>, 'DROP': <TokenType.DROP: 'DROP'>, 'ELSE': <TokenType.ELSE: 'ELSE'>, 'END': <TokenType.END: 'END'>, 'ESCAPE': <TokenType.ESCAPE: 'ESCAPE'>, 'EXCEPT': <TokenType.EXCEPT: 'EXCEPT'>, 'EXECUTE': <TokenType.EXECUTE: 'EXECUTE'>, 'EXISTS': <TokenType.EXISTS: 'EXISTS'>, 'FALSE': <TokenType.FALSE: 'FALSE'>, 'FETCH': <TokenType.FETCH: 'FETCH'>, 'FILTER': <TokenType.FILTER: 'FILTER'>, 'FIRST': <TokenType.FIRST: 'FIRST'>, 'FULL': <TokenType.FULL: 'FULL'>, 'FUNCTION': <TokenType.FUNCTION: 'FUNCTION'>, 'FOR': <TokenType.FOR: 'FOR'>, 'FOREIGN KEY': <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, 'FORMAT': <TokenType.FORMAT: 'FORMAT'>, 'FROM': <TokenType.FROM: 'FROM'>, 'GEOGRAPHY': <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, 'GEOMETRY': <TokenType.GEOMETRY: 'GEOMETRY'>, 'GLOB': <TokenType.GLOB: 'GLOB'>, 'GROUP BY': <TokenType.GROUP_BY: 'GROUP_BY'>, 'GROUPING SETS': <TokenType.GROUPING_SETS: 'GROUPING_SETS'>, 'HAVING': <TokenType.HAVING: 'HAVING'>, 'ILIKE': <TokenType.ILIKE: 'ILIKE'>, 'IN': <TokenType.IN: 'IN'>, 'INDEX': <TokenType.INDEX: 'INDEX'>, 'INET': <TokenType.INET: 'INET'>, 'INNER': <TokenType.INNER: 'INNER'>, 'INSERT': <TokenType.INSERT: 'INSERT'>, 'INTERVAL': <TokenType.INTERVAL: 'INTERVAL'>, 'INTERSECT': <TokenType.INTERSECT: 'INTERSECT'>, 'INTO': <TokenType.INTO: 'INTO'>, 'IS': <TokenType.IS: 'IS'>, 'ISNULL': <TokenType.ISNULL: 'ISNULL'>, 'JOIN': <TokenType.JOIN: 'JOIN'>, 'KEEP': <TokenType.KEEP: 'KEEP'>, 'KILL': <TokenType.KILL: 'KILL'>, 'LATERAL': <TokenType.LATERAL: 'LATERAL'>, 'LEFT': <TokenType.LEFT: 'LEFT'>, 'LIKE': <TokenType.LIKE: 'LIKE'>, 'LIMIT': <TokenType.LIMIT: 'LIMIT'>, 'LOAD': <TokenType.LOAD: 'LOAD'>, 'LOCK': <TokenType.LOCK: 'LOCK'>, 'MERGE': <TokenType.MERGE: 'MERGE'>, 'NATURAL': <TokenType.NATURAL: 'NATURAL'>, 'NEXT': <TokenType.NEXT: 'NEXT'>, 'NOT': <TokenType.NOT: 'NOT'>, 'NOTNULL': <TokenType.NOTNULL: 'NOTNULL'>, 'NULL': <TokenType.NULL: 'NULL'>, 'OBJECT': <TokenType.OBJECT: 'OBJECT'>, 'OFFSET': <TokenType.OFFSET: 'OFFSET'>, 'ON': <TokenType.ON: 'ON'>, 'OR': <TokenType.OR: 'OR'>, 'XOR': <TokenType.XOR: 'XOR'>, 'ORDER BY': <TokenType.ORDER_BY: 'ORDER_BY'>, 'ORDINALITY': <TokenType.ORDINALITY: 'ORDINALITY'>, 'OUTER': <TokenType.OUTER: 'OUTER'>, 'OVER': <TokenType.OVER: 'OVER'>, 'OVERLAPS': <TokenType.OVERLAPS: 'OVERLAPS'>, 'OVERWRITE': <TokenType.OVERWRITE: 'OVERWRITE'>, 'PARTITION': <TokenType.PARTITION: 'PARTITION'>, 'PARTITION BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED_BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PERCENT': <TokenType.PERCENT: 'PERCENT'>, 'PIVOT': <TokenType.PIVOT: 'PIVOT'>, 'PRAGMA': <TokenType.PRAGMA: 'PRAGMA'>, 'PRIMARY KEY': <TokenType.PRIMARY_KEY: 'PRIMARY_KEY'>, 'PROCEDURE': <TokenType.PROCEDURE: 'PROCEDURE'>, 'QUALIFY': <TokenType.QUALIFY: 'QUALIFY'>, 'RANGE': <TokenType.RANGE: 'RANGE'>, 'RECURSIVE': <TokenType.RECURSIVE: 'RECURSIVE'>, 'REGEXP': <TokenType.RLIKE: 'RLIKE'>, 'REPLACE': <TokenType.REPLACE: 'REPLACE'>, 'RETURNING': <TokenType.RETURNING: 'RETURNING'>, 'REFERENCES': <TokenType.REFERENCES: 'REFERENCES'>, 'RIGHT': <TokenType.RIGHT: 'RIGHT'>, 'RLIKE': <TokenType.RLIKE: 'RLIKE'>, 'ROLLBACK': <TokenType.ROLLBACK: 'ROLLBACK'>, 'ROLLUP': <TokenType.ROLLUP: 'ROLLUP'>, 'ROW': <TokenType.ROW: 'ROW'>, 'ROWS': <TokenType.ROWS: 'ROWS'>, 'SCHEMA': <TokenType.SCHEMA: 'SCHEMA'>, 'SELECT': <TokenType.SELECT: 'SELECT'>, 'SEMI': <TokenType.SEMI: 'SEMI'>, 'SET': <TokenType.SET: 'SET'>, 'SETTINGS': <TokenType.SETTINGS: 'SETTINGS'>, 'SHOW': <TokenType.SHOW: 'SHOW'>, 'SIMILAR TO': <TokenType.SIMILAR_TO: 'SIMILAR_TO'>, 'SOME': <TokenType.SOME: 'SOME'>, 'SORT BY': <TokenType.SORT_BY: 'SORT_BY'>, 'START WITH': <TokenType.START_WITH: 'START_WITH'>, 'TABLE': <TokenType.TABLE: 'TABLE'>, 'TABLESAMPLE': <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>, 'TEMP': <TokenType.TEMPORARY: 'TEMPORARY'>, 'TEMPORARY': <TokenType.TEMPORARY: 'TEMPORARY'>, 'THEN': <TokenType.THEN: 'THEN'>, 'TRUE': <TokenType.TRUE: 'TRUE'>, 'UNION': <TokenType.UNION: 'UNION'>, 'UNKNOWN': <TokenType.UNKNOWN: 'UNKNOWN'>, 'UNNEST': <TokenType.UNNEST: 'UNNEST'>, 'UNPIVOT': <TokenType.UNPIVOT: 'UNPIVOT'>, 'UPDATE': <TokenType.UPDATE: 'UPDATE'>, 'USE': <TokenType.USE: 'USE'>, 'USING': <TokenType.USING: 'USING'>, 'UUID': <TokenType.UUID: 'UUID'>, 'VALUES': <TokenType.VALUES: 'VALUES'>, 'VIEW': <TokenType.VIEW: 'VIEW'>, 'VOLATILE': <TokenType.VOLATILE: 'VOLATILE'>, 'WHEN': <TokenType.WHEN: 'WHEN'>, 'WHERE': <TokenType.WHERE: 'WHERE'>, 'WINDOW': <TokenType.WINDOW: 'WINDOW'>, 'WITH': <TokenType.WITH: 'WITH'>, 'APPLY': <TokenType.APPLY: 'APPLY'>, 'ARRAY': <TokenType.ARRAY: 'ARRAY'>, 'BIT': <TokenType.BIT: 'BIT'>, 'BOOL': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BOOLEAN': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BYTE': <TokenType.TINYINT: 'TINYINT'>, 'MEDIUMINT': <TokenType.MEDIUMINT: 'MEDIUMINT'>, 'INT1': <TokenType.TINYINT: 'TINYINT'>, 'TINYINT': <TokenType.TINYINT: 'TINYINT'>, 'INT16': <TokenType.SMALLINT: 'SMALLINT'>, 'SHORT': <TokenType.SMALLINT: 'SMALLINT'>, 'SMALLINT': <TokenType.SMALLINT: 'SMALLINT'>, 'INT128': <TokenType.INT128: 'INT128'>, 'HUGEINT': <TokenType.INT128: 'INT128'>, 'INT2': <TokenType.SMALLINT: 'SMALLINT'>, 'INTEGER': <TokenType.INT: 'INT'>, 'INT': <TokenType.INT: 'INT'>, 'INT4': <TokenType.INT: 'INT'>, 'INT32': <TokenType.INT: 'INT'>, 'INT64': <TokenType.BIGINT: 'BIGINT'>, 'LONG': <TokenType.BIGINT: 'BIGINT'>, 'BIGINT': <TokenType.BIGINT: 'BIGINT'>, 'INT8': <TokenType.TINYINT: 'TINYINT'>, 'DEC': <TokenType.DECIMAL: 'DECIMAL'>, 'DECIMAL': <TokenType.DECIMAL: 'DECIMAL'>, 'BIGDECIMAL': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'BIGNUMERIC': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'MAP': <TokenType.MAP: 'MAP'>, 'NULLABLE': <TokenType.NULLABLE: 'NULLABLE'>, 'NUMBER': <TokenType.DECIMAL: 'DECIMAL'>, 'NUMERIC': <TokenType.DECIMAL: 'DECIMAL'>, 'FIXED': <TokenType.DECIMAL: 'DECIMAL'>, 'REAL': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT4': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT8': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE PRECISION': <TokenType.DOUBLE: 'DOUBLE'>, 'JSON': <TokenType.JSON: 'JSON'>, 'CHAR': <TokenType.CHAR: 'CHAR'>, 'CHARACTER': <TokenType.CHAR: 'CHAR'>, 'NCHAR': <TokenType.NCHAR: 'NCHAR'>, 'VARCHAR': <TokenType.VARCHAR: 'VARCHAR'>, 'VARCHAR2': <TokenType.VARCHAR: 'VARCHAR'>, 'NVARCHAR': <TokenType.NVARCHAR: 'NVARCHAR'>, 'NVARCHAR2': <TokenType.NVARCHAR: 'NVARCHAR'>, 'STR': <TokenType.TEXT: 'TEXT'>, 'STRING': <TokenType.TEXT: 'TEXT'>, 'TEXT': <TokenType.TEXT: 'TEXT'>, 'LONGTEXT': <TokenType.LONGTEXT: 'LONGTEXT'>, 'MEDIUMTEXT': <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, 'TINYTEXT': <TokenType.TINYTEXT: 'TINYTEXT'>, 'CLOB': <TokenType.TEXT: 'TEXT'>, 'LONGVARCHAR': <TokenType.TEXT: 'TEXT'>, 'BINARY': <TokenType.BINARY: 'BINARY'>, 'BLOB': <TokenType.VARBINARY: 'VARBINARY'>, 'LONGBLOB': <TokenType.LONGBLOB: 'LONGBLOB'>, 'MEDIUMBLOB': <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, 'TINYBLOB': <TokenType.TINYBLOB: 'TINYBLOB'>, 'BYTEA': <TokenType.VARBINARY: 'VARBINARY'>, 'VARBINARY': <TokenType.VARBINARY: 'VARBINARY'>, 'TIME': <TokenType.TIME: 'TIME'>, 'TIMETZ': <TokenType.TIMETZ: 'TIMETZ'>, 'TIMESTAMP': <TokenType.TIMESTAMP: 'TIMESTAMP'>, 'TIMESTAMPTZ': <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, 'TIMESTAMPLTZ': <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, 'DATE': <TokenType.DATE: 'DATE'>, 'DATETIME': <TokenType.DATETIME: 'DATETIME'>, 'INT4RANGE': <TokenType.INT4RANGE: 'INT4RANGE'>, 'INT4MULTIRANGE': <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, 'INT8RANGE': <TokenType.INT8RANGE: 'INT8RANGE'>, 'INT8MULTIRANGE': <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, 'NUMRANGE': <TokenType.NUMRANGE: 'NUMRANGE'>, 'NUMMULTIRANGE': <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, 'TSRANGE': <TokenType.TSRANGE: 'TSRANGE'>, 'TSMULTIRANGE': <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, 'TSTZRANGE': <TokenType.TSTZRANGE: 'TSTZRANGE'>, 'TSTZMULTIRANGE': <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, 'DATERANGE': <TokenType.DATERANGE: 'DATERANGE'>, 'DATEMULTIRANGE': <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, 'UNIQUE': <TokenType.UNIQUE: 'UNIQUE'>, 'STRUCT': <TokenType.STRUCT: 'STRUCT'>, 'VARIANT': <TokenType.VARIANT: 'VARIANT'>, 'ALTER': <TokenType.ALTER: 'ALTER'>, 'ANALYZE': <TokenType.COMMAND: 'COMMAND'>, 'CALL': <TokenType.COMMAND: 'COMMAND'>, 'COMMENT': <TokenType.COMMENT: 'COMMENT'>, 'COPY': <TokenType.COMMAND: 'COMMAND'>, 'EXPLAIN': <TokenType.COMMAND: 'COMMAND'>, 'GRANT': <TokenType.COMMAND: 'COMMAND'>, 'OPTIMIZE': <TokenType.COMMAND: 'COMMAND'>, 'PREPARE': <TokenType.COMMAND: 'COMMAND'>, 'TRUNCATE': <TokenType.COMMAND: 'COMMAND'>, 'VACUUM': <TokenType.COMMAND: 'COMMAND'>, 'USER-DEFINED': <TokenType.USERDEFINED: 'USERDEFINED'>, 'FOR VERSION': <TokenType.VERSION_SNAPSHOT: 'VERSION_SNAPSHOT'>, 'FOR TIMESTAMP': <TokenType.TIMESTAMP_SNAPSHOT: 'TIMESTAMP_SNAPSHOT'>, 'ATTACH': <TokenType.COMMAND: 'COMMAND'>, 'DATETIME64': <TokenType.DATETIME64: 'DATETIME64'>, 'DICTIONARY': <TokenType.DICTIONARY: 'DICTIONARY'>, 'ENUM': <TokenType.ENUM: 'ENUM'>, 'ENUM8': <TokenType.ENUM8: 'ENUM8'>, 'ENUM16': <TokenType.ENUM16: 'ENUM16'>, 'FINAL': <TokenType.FINAL: 'FINAL'>, 'FIXEDSTRING': <TokenType.FIXEDSTRING: 'FIXEDSTRING'>, 'FLOAT32': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT64': <TokenType.DOUBLE: 'DOUBLE'>, 'GLOBAL': <TokenType.GLOBAL: 'GLOBAL'>, 'INT256': <TokenType.INT256: 'INT256'>, 'LOWCARDINALITY': <TokenType.LOWCARDINALITY: 'LOWCARDINALITY'>, 'NESTED': <TokenType.NESTED: 'NESTED'>, 'SAMPLE': <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>, 'TUPLE': <TokenType.STRUCT: 'STRUCT'>, 'UINT128': <TokenType.UINT128: 'UINT128'>, 'UINT16': <TokenType.USMALLINT: 'USMALLINT'>, 'UINT256': <TokenType.UINT256: 'UINT256'>, 'UINT32': <TokenType.UINT: 'UINT'>, 'UINT64': <TokenType.UBIGINT: 'UBIGINT'>, 'UINT8': <TokenType.UTINYINT: 'UTINYINT'>}
SINGLE_TOKENS = {'(': <TokenType.L_PAREN: 'L_PAREN'>, ')': <TokenType.R_PAREN: 'R_PAREN'>, '[': <TokenType.L_BRACKET: 'L_BRACKET'>, ']': <TokenType.R_BRACKET: 'R_BRACKET'>, '{': <TokenType.L_BRACE: 'L_BRACE'>, '}': <TokenType.R_BRACE: 'R_BRACE'>, '&': <TokenType.AMP: 'AMP'>, '^': <TokenType.CARET: 'CARET'>, ':': <TokenType.COLON: 'COLON'>, ',': <TokenType.COMMA: 'COMMA'>, '.': <TokenType.DOT: 'DOT'>, '-': <TokenType.DASH: 'DASH'>, '=': <TokenType.EQ: 'EQ'>, '>': <TokenType.GT: 'GT'>, '<': <TokenType.LT: 'LT'>, '%': <TokenType.MOD: 'MOD'>, '!': <TokenType.NOT: 'NOT'>, '|': <TokenType.PIPE: 'PIPE'>, '+': <TokenType.PLUS: 'PLUS'>, ';': <TokenType.SEMICOLON: 'SEMICOLON'>, '/': <TokenType.SLASH: 'SLASH'>, '\\': <TokenType.BACKSLASH: 'BACKSLASH'>, '*': <TokenType.STAR: 'STAR'>, '~': <TokenType.TILDA: 'TILDA'>, '?': <TokenType.PLACEHOLDER: 'PLACEHOLDER'>, '@': <TokenType.PARAMETER: 'PARAMETER'>, "'": <TokenType.QUOTE: 'QUOTE'>, '`': <TokenType.IDENTIFIER: 'IDENTIFIER'>, '"': <TokenType.IDENTIFIER: 'IDENTIFIER'>, '#': <TokenType.HASH: 'HASH'>, '$': <TokenType.HEREDOC_STRING: 'HEREDOC_STRING'>}

##### Inherited Members

sqlglot.tokens.Tokenizer
  Tokenizer
  BYTE_STRINGS
  RAW_STRINGS
  IDENTIFIER_ESCAPES
  QUOTES
  VAR_SINGLE_TOKENS
  WHITE_SPACE
  COMMANDS
  COMMAND_PREFIX_TOKENS
  NUMERIC_LITERALS
  dialect
  reset
  tokenize
  peek
  tokenize_rs
  size
  sql
  tokens

class ClickHouse.Parser(sqlglot.parser.Parser): View Source

     86    class Parser(parser.Parser):
     87        # Tested in ClickHouse's playground, it seems that the following two queries do the same thing
     88        # * select x from t1 union all select x from t2 limit 1;
     89        # * select x from t1 union all (select x from t2 limit 1);
     90        MODIFIERS_ATTACHED_TO_UNION = False
     91
     92        FUNCTIONS = {
     93            **parser.Parser.FUNCTIONS,
     94            "ANY": exp.AnyValue.from_arg_list,
     95            "DATE_ADD": lambda args: exp.DateAdd(
     96                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
     97            ),
     98            "DATEADD": lambda args: exp.DateAdd(
     99                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    100            ),
    101            "DATE_DIFF": lambda args: exp.DateDiff(
    102                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    103            ),
    104            "DATEDIFF": lambda args: exp.DateDiff(
    105                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    106            ),
    107            "MAP": parse_var_map,
    108            "MATCH": exp.RegexpLike.from_arg_list,
    109            "UNIQ": exp.ApproxDistinct.from_arg_list,
    110            "XOR": lambda args: exp.Xor(expressions=args),
    111        }
    112
    113        FUNCTIONS_WITH_ALIASED_ARGS = {*parser.Parser.FUNCTIONS_WITH_ALIASED_ARGS, "TUPLE"}
    114
    115        FUNCTION_PARSERS = {
    116            **parser.Parser.FUNCTION_PARSERS,
    117            "ARRAYJOIN": lambda self: self.expression(exp.Explode, this=self._parse_expression()),
    118            "QUANTILE": lambda self: self._parse_quantile(),
    119        }
    120
    121        FUNCTION_PARSERS.pop("MATCH")
    122
    123        NO_PAREN_FUNCTION_PARSERS = parser.Parser.NO_PAREN_FUNCTION_PARSERS.copy()
    124        NO_PAREN_FUNCTION_PARSERS.pop("ANY")
    125
    126        RANGE_PARSERS = {
    127            **parser.Parser.RANGE_PARSERS,
    128            TokenType.GLOBAL: lambda self, this: self._match(TokenType.IN)
    129            and self._parse_in(this, is_global=True),
    130        }
    131
    132        # The PLACEHOLDER entry is popped because 1) it doesn't affect Clickhouse (it corresponds to
    133        # the postgres-specific JSONBContains parser) and 2) it makes parsing the ternary op simpler.
    134        COLUMN_OPERATORS = parser.Parser.COLUMN_OPERATORS.copy()
    135        COLUMN_OPERATORS.pop(TokenType.PLACEHOLDER)
    136
    137        JOIN_KINDS = {
    138            *parser.Parser.JOIN_KINDS,
    139            TokenType.ANY,
    140            TokenType.ASOF,
    141            TokenType.ARRAY,
    142        }
    143
    144        TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS - {
    145            TokenType.ANY,
    146            TokenType.SETTINGS,
    147            TokenType.FORMAT,
    148            TokenType.ARRAY,
    149        }
    150
    151        LOG_DEFAULTS_TO_LN = True
    152
    153        QUERY_MODIFIER_PARSERS = {
    154            **parser.Parser.QUERY_MODIFIER_PARSERS,
    155            TokenType.SETTINGS: lambda self: (
    156                "settings",
    157                self._advance() or self._parse_csv(self._parse_conjunction),
    158            ),
    159            TokenType.FORMAT: lambda self: ("format", self._advance() or self._parse_id_var()),
    160        }
    161
    162        def _parse_conjunction(self) -> t.Optional[exp.Expression]:
    163            this = super()._parse_conjunction()
    164
    165            if self._match(TokenType.PLACEHOLDER):
    166                return self.expression(
    167                    exp.If,
    168                    this=this,
    169                    true=self._parse_conjunction(),
    170                    false=self._match(TokenType.COLON) and self._parse_conjunction(),
    171                )
    172
    173            return this
    174
    175        def _parse_placeholder(self) -> t.Optional[exp.Expression]:
    176            """
    177            Parse a placeholder expression like SELECT {abc: UInt32} or FROM {table: Identifier}
    178            https://clickhouse.com/docs/en/sql-reference/syntax#defining-and-using-query-parameters
    179            """
    180            if not self._match(TokenType.L_BRACE):
    181                return None
    182
    183            this = self._parse_id_var()
    184            self._match(TokenType.COLON)
    185            kind = self._parse_types(check_func=False, allow_identifiers=False) or (
    186                self._match_text_seq("IDENTIFIER") and "Identifier"
    187            )
    188
    189            if not kind:
    190                self.raise_error("Expecting a placeholder type or 'Identifier' for tables")
    191            elif not self._match(TokenType.R_BRACE):
    192                self.raise_error("Expecting }")
    193
    194            return self.expression(exp.Placeholder, this=this, kind=kind)
    195
    196        def _parse_in(self, this: t.Optional[exp.Expression], is_global: bool = False) -> exp.In:
    197            this = super()._parse_in(this)
    198            this.set("is_global", is_global)
    199            return this
    200
    201        def _parse_table(
    202            self,
    203            schema: bool = False,
    204            joins: bool = False,
    205            alias_tokens: t.Optional[t.Collection[TokenType]] = None,
    206            parse_bracket: bool = False,
    207        ) -> t.Optional[exp.Expression]:
    208            this = super()._parse_table(
    209                schema=schema, joins=joins, alias_tokens=alias_tokens, parse_bracket=parse_bracket
    210            )
    211
    212            if self._match(TokenType.FINAL):
    213                this = self.expression(exp.Final, this=this)
    214
    215            return this
    216
    217        def _parse_position(self, haystack_first: bool = False) -> exp.StrPosition:
    218            return super()._parse_position(haystack_first=True)
    219
    220        # https://clickhouse.com/docs/en/sql-reference/statements/select/with/
    221        def _parse_cte(self) -> exp.CTE:
    222            index = self._index
    223            try:
    224                # WITH <identifier> AS <subquery expression>
    225                return super()._parse_cte()
    226            except ParseError:
    227                # WITH <expression> AS <identifier>
    228                self._retreat(index)
    229
    230                return self.expression(
    231                    exp.CTE,
    232                    this=self._parse_field(),
    233                    alias=self._parse_table_alias(),
    234                    scalar=True,
    235                )
    236
    237        def _parse_join_parts(
    238            self,
    239        ) -> t.Tuple[t.Optional[Token], t.Optional[Token], t.Optional[Token]]:
    240            is_global = self._match(TokenType.GLOBAL) and self._prev
    241            kind_pre = self._match_set(self.JOIN_KINDS, advance=False) and self._prev
    242
    243            if kind_pre:
    244                kind = self._match_set(self.JOIN_KINDS) and self._prev
    245                side = self._match_set(self.JOIN_SIDES) and self._prev
    246                return is_global, side, kind
    247
    248            return (
    249                is_global,
    250                self._match_set(self.JOIN_SIDES) and self._prev,
    251                self._match_set(self.JOIN_KINDS) and self._prev,
    252            )
    253
    254        def _parse_join(
    255            self, skip_join_token: bool = False, parse_bracket: bool = False
    256        ) -> t.Optional[exp.Join]:
    257            join = super()._parse_join(skip_join_token=skip_join_token, parse_bracket=True)
    258
    259            if join:
    260                join.set("global", join.args.pop("method", None))
    261            return join
    262
    263        def _parse_function(
    264            self,
    265            functions: t.Optional[t.Dict[str, t.Callable]] = None,
    266            anonymous: bool = False,
    267            optional_parens: bool = True,
    268        ) -> t.Optional[exp.Expression]:
    269            func = super()._parse_function(
    270                functions=functions, anonymous=anonymous, optional_parens=optional_parens
    271            )
    272
    273            if isinstance(func, exp.Anonymous):
    274                params = self._parse_func_params(func)
    275
    276                if params:
    277                    return self.expression(
    278                        exp.ParameterizedAgg,
    279                        this=func.this,
    280                        expressions=func.expressions,
    281                        params=params,
    282                    )
    283
    284            return func
    285
    286        def _parse_func_params(
    287            self, this: t.Optional[exp.Func] = None
    288        ) -> t.Optional[t.List[exp.Expression]]:
    289            if self._match_pair(TokenType.R_PAREN, TokenType.L_PAREN):
    290                return self._parse_csv(self._parse_lambda)
    291
    292            if self._match(TokenType.L_PAREN):
    293                params = self._parse_csv(self._parse_lambda)
    294                self._match_r_paren(this)
    295                return params
    296
    297            return None
    298
    299        def _parse_quantile(self) -> exp.Quantile:
    300            this = self._parse_lambda()
    301            params = self._parse_func_params()
    302            if params:
    303                return self.expression(exp.Quantile, this=params[0], quantile=this)
    304            return self.expression(exp.Quantile, this=this, quantile=exp.Literal.number(0.5))
    305
    306        def _parse_wrapped_id_vars(self, optional: bool = False) -> t.List[exp.Expression]:
    307            return super()._parse_wrapped_id_vars(optional=True)
    308
    309        def _parse_primary_key(
    310            self, wrapped_optional: bool = False, in_props: bool = False
    311        ) -> exp.PrimaryKeyColumnConstraint | exp.PrimaryKey:
    312            return super()._parse_primary_key(
    313                wrapped_optional=wrapped_optional or in_props, in_props=in_props
    314            )
    315
    316        def _parse_on_property(self) -> t.Optional[exp.Expression]:
    317            index = self._index
    318            if self._match_text_seq("CLUSTER"):
    319                this = self._parse_id_var()
    320                if this:
    321                    return self.expression(exp.OnCluster, this=this)
    322                else:
    323                    self._retreat(index)
    324            return None

Parser consumes a list of tokens produced by the Tokenizer and produces a parsed syntax tree.

###### Arguments:

- error_level: The desired error level. Default: ErrorLevel.IMMEDIATE
- error_message_context: Determines the amount of context to capture from a query string when displaying the error message (in number of characters). Default: 100
- max_errors: Maximum number of error messages to include in a raised ParseError. This is only relevant if error_level is ErrorLevel.RAISE. Default: 3

MODIFIERS_ATTACHED_TO_UNION = False
FUNCTIONS = {'ABS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Abs'>>, 'ANY_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnyValue'>>, 'APPROX_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_COUNT_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxQuantile'>>, 'APPROX_TOP_K': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxTopK'>>, 'ARG_MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARGMAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'MAX_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARG_MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARGMIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'MIN_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Array'>>, 'ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAgg'>>, 'ARRAY_ALL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAll'>>, 'ARRAY_ANY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAny'>>, 'ARRAY_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_JOIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayJoin'>>, 'ARRAY_SIZE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_SORT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySort'>>, 'ARRAY_SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySum'>>, 'ARRAY_UNION_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUnionAgg'>>, 'ARRAY_UNIQUE_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUniqueAgg'>>, 'AVG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Avg'>>, 'CASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Case'>>, 'CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cast'>>, 'CAST_TO_STR_TYPE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CastToStrType'>>, 'CEIL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CEILING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CHR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'COALESCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'IFNULL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'NVL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'COLLATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Collate'>>, 'CONCAT': <function Parser.<lambda>>, 'CONCAT_WS': <function Parser.<lambda>>, 'COUNT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Count'>>, 'COUNT_IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'CURRENT_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDate'>>, 'CURRENT_DATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDatetime'>>, 'CURRENT_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTime'>>, 'CURRENT_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'CURRENT_USER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Date'>>, 'DATE_ADD': <function ClickHouse.Parser.<lambda>>, 'DATEDIFF': <function ClickHouse.Parser.<lambda>>, 'DATE_DIFF': <function ClickHouse.Parser.<lambda>>, 'DATEFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATE_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateStrToDate'>>, 'DATE_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateSub'>>, 'DATE_TO_DATE_STR': <function Parser.<lambda>>, 'DATE_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateToDi'>>, 'DATE_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateTrunc'>>, 'DATETIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeAdd'>>, 'DATETIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeDiff'>>, 'DATETIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeSub'>>, 'DATETIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeTrunc'>>, 'DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Day'>>, 'DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAYOFMONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAY_OF_WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAYOFWEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAY_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DAYOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DECODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Decode'>>, 'DI_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DiToDate'>>, 'ENCODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Encode'>>, 'EXP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Exp'>>, 'EXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Explode'>>, 'EXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ExplodeOuter'>>, 'EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Extract'>>, 'FIRST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.First'>>, 'FLATTEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Flatten'>>, 'FLOOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Floor'>>, 'FROM_BASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase'>>, 'FROM_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase64'>>, 'GENERATE_SERIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateSeries'>>, 'GREATEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Greatest'>>, 'GROUP_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GroupConcat'>>, 'HEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hex'>>, 'HLL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hll'>>, 'IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'INITCAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Initcap'>>, 'IS_INF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'ISINF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'IS_NAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'ISNAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'J_S_O_N_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArray'>>, 'J_S_O_N_ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayAgg'>>, 'JSON_ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayContains'>>, 'JSONB_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtract'>>, 'JSONB_EXTRACT_SCALAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtractScalar'>>, 'JSON_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONExtract'>>, 'JSON_EXTRACT_SCALAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONExtractScalar'>>, 'JSON_FORMAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONFormat'>>, 'J_S_O_N_OBJECT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObject'>>, 'J_S_O_N_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONTable'>>, 'LAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Last'>>, 'LAST_DATE_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDateOfMonth'>>, 'LEAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Least'>>, 'LEFT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Left'>>, 'LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEVENSHTEIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Levenshtein'>>, 'LN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ln'>>, 'LOG': <function parse_logarithm>, 'LOG10': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Log10'>>, 'LOG2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Log2'>>, 'LOGICAL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOLAND_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'LOGICAL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOLOR_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'LOWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'LCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'MD5': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5'>>, 'MD5_DIGEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5Digest'>>, 'MAP': <function parse_var_map>, 'MAP_FROM_ENTRIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MapFromEntries'>>, 'MATCH_AGAINST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MatchAgainst'>>, 'MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Max'>>, 'MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Min'>>, 'MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Month'>>, 'MONTHS_BETWEEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MonthsBetween'>>, 'NEXT_VALUE_FOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NextValueFor'>>, 'NULLIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nullif'>>, 'NUMBER_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NumberToStr'>>, 'NVL2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nvl2'>>, 'OPEN_J_S_O_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.OpenJSON'>>, 'PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParameterizedAgg'>>, 'PARSE_JSON': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'JSON_PARSE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'PERCENTILE_CONT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileCont'>>, 'PERCENTILE_DISC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileDisc'>>, 'POSEXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Posexplode'>>, 'POSEXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PosexplodeOuter'>>, 'POWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'POW': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'PREDICT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Predict'>>, 'QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Quantile'>>, 'RANGE_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RangeN'>>, 'READ_CSV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ReadCSV'>>, 'REDUCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Reduce'>>, 'REGEXP_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpExtract'>>, 'REGEXP_I_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpILike'>>, 'REGEXP_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpLike'>>, 'REGEXP_REPLACE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpReplace'>>, 'REGEXP_SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpSplit'>>, 'REPEAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Repeat'>>, 'RIGHT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Right'>>, 'ROUND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Round'>>, 'ROW_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RowNumber'>>, 'SHA': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA1': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA2'>>, 'SAFE_DIVIDE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SafeDivide'>>, 'SORT_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SortArray'>>, 'SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'SQRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sqrt'>>, 'STANDARD_HASH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StandardHash'>>, 'STAR_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StarMap'>>, 'STARTS_WITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STARTSWITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STDDEV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stddev'>>, 'STDDEV_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevPop'>>, 'STDDEV_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevSamp'>>, 'STR_POSITION': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrPosition'>>, 'STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToDate'>>, 'STR_TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToMap'>>, 'STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToTime'>>, 'STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToUnix'>>, 'STRUCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Struct'>>, 'STRUCT_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StructExtract'>>, 'STUFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'INSERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'SUBSTRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Substring'>>, 'SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sum'>>, 'TIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeAdd'>>, 'TIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeDiff'>>, 'TIME_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToDate'>>, 'TIME_STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToTime'>>, 'TIME_STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToUnix'>>, 'TIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeSub'>>, 'TIME_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToStr'>>, 'TIME_TO_TIME_STR': <function Parser.<lambda>>, 'TIME_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToUnix'>>, 'TIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeTrunc'>>, 'TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Timestamp'>>, 'TIMESTAMP_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampAdd'>>, 'TIMESTAMP_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMP_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampSub'>>, 'TIMESTAMP_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampTrunc'>>, 'TO_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToBase64'>>, 'TO_CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToChar'>>, 'TO_DAYS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToDays'>>, 'TRANSFORM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Transform'>>, 'TRIM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Trim'>>, 'TRY_CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TryCast'>>, 'TS_OR_DI_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDiToDi'>>, 'TS_OR_DS_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsAdd'>>, 'TS_OR_DS_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsDiff'>>, 'TS_OR_DS_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToDate'>>, 'TS_OR_DS_TO_DATE_STR': <function Parser.<lambda>>, 'UNHEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Unhex'>>, 'UNIX_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToStr'>>, 'UNIX_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTime'>>, 'UNIX_TO_TIME_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTimeStr'>>, 'UPPER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'UCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'VAR_MAP': <function parse_var_map>, 'VARIANCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'VAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Week'>>, 'WEEK_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WEEKOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WHEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.When'>>, 'X_M_L_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.XMLTable'>>, 'XOR': <function ClickHouse.Parser.<lambda>>, 'YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Year'>>, 'GLOB': <function Parser.<lambda>>, 'LIKE': <function parse_like>, 'ANY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnyValue'>>, 'DATEADD': <function ClickHouse.Parser.<lambda>>, 'MATCH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpLike'>>, 'UNIQ': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>}
FUNCTIONS_WITH_ALIASED_ARGS = {'TUPLE', 'STRUCT'}
FUNCTION_PARSERS = {'ANY_VALUE': <function Parser.<lambda>>, 'CAST': <function Parser.<lambda>>, 'CONVERT': <function Parser.<lambda>>, 'DECODE': <function Parser.<lambda>>, 'EXTRACT': <function Parser.<lambda>>, 'JSON_OBJECT': <function Parser.<lambda>>, 'JSON_TABLE': <function Parser.<lambda>>, 'OPENJSON': <function Parser.<lambda>>, 'POSITION': <function Parser.<lambda>>, 'PREDICT': <function Parser.<lambda>>, 'SAFE_CAST': <function Parser.<lambda>>, 'STRING_AGG': <function Parser.<lambda>>, 'SUBSTRING': <function Parser.<lambda>>, 'TRIM': <function Parser.<lambda>>, 'TRY_CAST': <function Parser.<lambda>>, 'TRY_CONVERT': <function Parser.<lambda>>, 'ARRAYJOIN': <function ClickHouse.Parser.<lambda>>, 'QUANTILE': <function ClickHouse.Parser.<lambda>>}
NO_PAREN_FUNCTION_PARSERS = {'CASE': <function Parser.<lambda>>, 'IF': <function Parser.<lambda>>, 'NEXT': <function Parser.<lambda>>}
RANGE_PARSERS = {<TokenType.BETWEEN: 'BETWEEN'>: <function Parser.<lambda>>, <TokenType.GLOB: 'GLOB'>: <function binary_range_parser.<locals>.<lambda>>, <TokenType.ILIKE: 'ILIKE'>: <function binary_range_parser.<locals>.<lambda>>, <TokenType.IN: 'IN'>: <function Parser.<lambda>>, <TokenType.IRLIKE: 'IRLIKE'>: <function binary_range_parser.<locals>.<lambda>>, <TokenType.IS: 'IS'>: <function Parser.<lambda>>, <TokenType.LIKE: 'LIKE'>: <function binary_range_parser.<locals>.<lambda>>, <TokenType.OVERLAPS: 'OVERLAPS'>: <function binary_range_parser.<locals>.<lambda>>, <TokenType.RLIKE: 'RLIKE'>: <function binary_range_parser.<locals>.<lambda>>, <TokenType.SIMILAR_TO: 'SIMILAR_TO'>: <function binary_range_parser.<locals>.<lambda>>, <TokenType.FOR: 'FOR'>: <function Parser.<lambda>>, <TokenType.GLOBAL: 'GLOBAL'>: <function ClickHouse.Parser.<lambda>>}
COLUMN_OPERATORS = {<TokenType.DOT: 'DOT'>: None, <TokenType.DCOLON: 'DCOLON'>: <function Parser.<lambda>>, <TokenType.ARROW: 'ARROW'>: <function Parser.<lambda>>, <TokenType.DARROW: 'DARROW'>: <function Parser.<lambda>>, <TokenType.HASH_ARROW: 'HASH_ARROW'>: <function Parser.<lambda>>, <TokenType.DHASH_ARROW: 'DHASH_ARROW'>: <function Parser.<lambda>>}
JOIN_KINDS = {<TokenType.ANY: 'ANY'>, <TokenType.SEMI: 'SEMI'>, <TokenType.ARRAY: 'ARRAY'>, <TokenType.INNER: 'INNER'>, <TokenType.ASOF: 'ASOF'>, <TokenType.OUTER: 'OUTER'>, <TokenType.CROSS: 'CROSS'>, <TokenType.ANTI: 'ANTI'>}
TABLE_ALIAS_TOKENS = {<TokenType.TIMESTAMP_MS: 'TIMESTAMP_MS'>, <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.USERDEFINED: 'USERDEFINED'>, <TokenType.VAR: 'VAR'>, <TokenType.TINYINT: 'TINYINT'>, <TokenType.DATE: 'DATE'>, <TokenType.CASE: 'CASE'>, <TokenType.MONEY: 'MONEY'>, <TokenType.HSTORE: 'HSTORE'>, <TokenType.SMALLMONEY: 'SMALLMONEY'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.TEXT: 'TEXT'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.UUID: 'UUID'>, <TokenType.INT4RANGE: 'INT4RANGE'>, <TokenType.TINYBLOB: 'TINYBLOB'>, <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, <TokenType.FIRST: 'FIRST'>, <TokenType.TIMESTAMP_S: 'TIMESTAMP_S'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.DOUBLE: 'DOUBLE'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.ROWS: 'ROWS'>, <TokenType.NUMRANGE: 'NUMRANGE'>, <TokenType.INDEX: 'INDEX'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.INT128: 'INT128'>, <TokenType.ALL: 'ALL'>, <TokenType.SET: 'SET'>, <TokenType.JSONB: 'JSONB'>, <TokenType.CACHE: 'CACHE'>, <TokenType.NULL: 'NULL'>, <TokenType.BOOLEAN: 'BOOLEAN'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, <TokenType.VARBINARY: 'VARBINARY'>, <TokenType.COMMENT: 'COMMENT'>, <TokenType.ROW: 'ROW'>, <TokenType.BIT: 'BIT'>, <TokenType.ENUM: 'ENUM'>, <TokenType.UDECIMAL: 'UDECIMAL'>, <TokenType.FALSE: 'FALSE'>, <TokenType.TIMESTAMP: 'TIMESTAMP'>, <TokenType.IMAGE: 'IMAGE'>, <TokenType.HLLSKETCH: 'HLLSKETCH'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.BIGINT: 'BIGINT'>, <TokenType.DATERANGE: 'DATERANGE'>, <TokenType.OBJECT_IDENTIFIER: 'OBJECT_IDENTIFIER'>, <TokenType.NEXT: 'NEXT'>, <TokenType.SOME: 'SOME'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.ASC: 'ASC'>, <TokenType.IPADDRESS: 'IPADDRESS'>, <TokenType.TINYTEXT: 'TINYTEXT'>, <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, <TokenType.XML: 'XML'>, <TokenType.LOAD: 'LOAD'>, <TokenType.TIMESTAMP_NS: 'TIMESTAMP_NS'>, <TokenType.TIMETZ: 'TIMETZ'>, <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, <TokenType.INTERVAL: 'INTERVAL'>, <TokenType.FIXEDSTRING: 'FIXEDSTRING'>, <TokenType.VARIANT: 'VARIANT'>, <TokenType.BINARY: 'BINARY'>, <TokenType.UMEDIUMINT: 'UMEDIUMINT'>, <TokenType.MERGE: 'MERGE'>, <TokenType.LONGBLOB: 'LONGBLOB'>, <TokenType.STRUCT: 'STRUCT'>, <TokenType.JSON: 'JSON'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.PSEUDO_TYPE: 'PSEUDO_TYPE'>, <TokenType.NESTED: 'NESTED'>, <TokenType.OBJECT: 'OBJECT'>, <TokenType.SMALLINT: 'SMALLINT'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.NULLABLE: 'NULLABLE'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.UINT: 'UINT'>, <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, <TokenType.SHOW: 'SHOW'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.USE: 'USE'>, <TokenType.GEOMETRY: 'GEOMETRY'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.MODEL: 'MODEL'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.UINT256: 'UINT256'>, <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, <TokenType.INT8RANGE: 'INT8RANGE'>, <TokenType.RANGE: 'RANGE'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.DESC: 'DESC'>, <TokenType.INET: 'INET'>, <TokenType.ENUM8: 'ENUM8'>, <TokenType.DECIMAL: 'DECIMAL'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.MAP: 'MAP'>, <TokenType.KILL: 'KILL'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.TIME: 'TIME'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, <TokenType.UBIGINT: 'UBIGINT'>, <TokenType.NVARCHAR: 'NVARCHAR'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.FLOAT: 'FLOAT'>, <TokenType.SERIAL: 'SERIAL'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.MEDIUMINT: 'MEDIUMINT'>, <TokenType.DIV: 'DIV'>, <TokenType.IS: 'IS'>, <TokenType.KEEP: 'KEEP'>, <TokenType.NCHAR: 'NCHAR'>, <TokenType.FILTER: 'FILTER'>, <TokenType.ENUM16: 'ENUM16'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.YEAR: 'YEAR'>, <TokenType.CHAR: 'CHAR'>, <TokenType.UTINYINT: 'UTINYINT'>, <TokenType.LOWCARDINALITY: 'LOWCARDINALITY'>, <TokenType.INT256: 'INT256'>, <TokenType.DATETIME: 'DATETIME'>, <TokenType.IPPREFIX: 'IPPREFIX'>, <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.UNKNOWN: 'UNKNOWN'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.TRUE: 'TRUE'>, <TokenType.LONGTEXT: 'LONGTEXT'>, <TokenType.BIGSERIAL: 'BIGSERIAL'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.END: 'END'>, <TokenType.TABLE: 'TABLE'>, <TokenType.UINT128: 'UINT128'>, <TokenType.TOP: 'TOP'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.DELETE: 'DELETE'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.INT: 'INT'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.SUPER: 'SUPER'>, <TokenType.TSTZRANGE: 'TSTZRANGE'>, <TokenType.DATETIME64: 'DATETIME64'>, <TokenType.VIEW: 'VIEW'>, <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, <TokenType.SMALLSERIAL: 'SMALLSERIAL'>, <TokenType.ROWVERSION: 'ROWVERSION'>, <TokenType.TSRANGE: 'TSRANGE'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.VARCHAR: 'VARCHAR'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.USMALLINT: 'USMALLINT'>}
LOG_DEFAULTS_TO_LN = True
QUERY_MODIFIER_PARSERS = {<TokenType.MATCH_RECOGNIZE: 'MATCH_RECOGNIZE'>: <function Parser.<lambda>>, <TokenType.WHERE: 'WHERE'>: <function Parser.<lambda>>, <TokenType.GROUP_BY: 'GROUP_BY'>: <function Parser.<lambda>>, <TokenType.HAVING: 'HAVING'>: <function Parser.<lambda>>, <TokenType.QUALIFY: 'QUALIFY'>: <function Parser.<lambda>>, <TokenType.WINDOW: 'WINDOW'>: <function Parser.<lambda>>, <TokenType.ORDER_BY: 'ORDER_BY'>: <function Parser.<lambda>>, <TokenType.LIMIT: 'LIMIT'>: <function Parser.<lambda>>, <TokenType.FETCH: 'FETCH'>: <function Parser.<lambda>>, <TokenType.OFFSET: 'OFFSET'>: <function Parser.<lambda>>, <TokenType.FOR: 'FOR'>: <function Parser.<lambda>>, <TokenType.LOCK: 'LOCK'>: <function Parser.<lambda>>, <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>: <function Parser.<lambda>>, <TokenType.USING: 'USING'>: <function Parser.<lambda>>, <TokenType.CLUSTER_BY: 'CLUSTER_BY'>: <function Parser.<lambda>>, <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>: <function Parser.<lambda>>, <TokenType.SORT_BY: 'SORT_BY'>: <function Parser.<lambda>>, <TokenType.CONNECT_BY: 'CONNECT_BY'>: <function Parser.<lambda>>, <TokenType.START_WITH: 'START_WITH'>: <function Parser.<lambda>>, <TokenType.SETTINGS: 'SETTINGS'>: <function ClickHouse.Parser.<lambda>>, <TokenType.FORMAT: 'FORMAT'>: <function ClickHouse.Parser.<lambda>>}
SHOW_TRIE: Dict = {}
SET_TRIE: Dict = {'GLOBAL': {0: True}, 'LOCAL': {0: True}, 'SESSION': {0: True}, 'TRANSACTION': {0: True}}

##### Inherited Members

sqlglot.parser.Parser
  Parser
  NO_PAREN_FUNCTIONS
  STRUCT_TYPE_TOKENS
  NESTED_TYPE_TOKENS
  ENUM_TYPE_TOKENS
  TYPE_TOKENS
  SIGNED_TO_UNSIGNED_TYPE_TOKEN
  SUBQUERY_PREDICATES
  RESERVED_TOKENS
  DB_CREATABLES
  CREATABLES
  ID_VAR_TOKENS
  INTERVAL_VARS
  COMMENT_TABLE_ALIAS_TOKENS
  UPDATE_ALIAS_TOKENS
  TRIM_TYPES
  FUNC_TOKENS
  CONJUNCTION
  EQUALITY
  COMPARISON
  BITWISE
  TERM
  FACTOR
  EXPONENT
  TIMES
  TIMESTAMPS
  SET_OPERATIONS
  JOIN_METHODS
  JOIN_SIDES
  JOIN_HINTS
  LAMBDAS
  EXPRESSION_PARSERS
  STATEMENT_PARSERS
  UNARY_PARSERS
  PRIMARY_PARSERS
  PLACEHOLDER_PARSERS
  PROPERTY_PARSERS
  CONSTRAINT_PARSERS
  ALTER_PARSERS
  SCHEMA_UNNAMED_CONSTRAINTS
  INVALID_FUNC_NAME_TOKENS
  SET_PARSERS
  SHOW_PARSERS
  TYPE_LITERAL_PARSERS
  MODIFIABLES
  DDL_SELECT_TOKENS
  PRE_VOLATILE_TOKENS
  TRANSACTION_KIND
  TRANSACTION_CHARACTERISTICS
  INSERT_ALTERNATIVES
  CLONE_KEYWORDS
  HISTORICAL_DATA_KIND
  OPCLASS_FOLLOW_KEYWORDS
  OPTYPE_FOLLOW_TOKENS
  TABLE_INDEX_HINT_TOKENS
  WINDOW_ALIAS_TOKENS
  WINDOW_BEFORE_PAREN_TOKENS
  WINDOW_SIDES
  FETCH_TOKENS
  ADD_CONSTRAINT_TOKENS
  DISTINCT_TOKENS
  NULL_TOKENS
  UNNEST_OFFSET_ALIAS_TOKENS
  STRICT_CAST
  PREFIXED_PIVOT_COLUMNS
  IDENTIFY_PIVOT_STRINGS
  ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN
  TABLESAMPLE_CSV
  SET_REQUIRES_ASSIGNMENT_DELIMITER
  TRIM_PATTERN_FIRST
  UNION_MODIFIERS
  error_level
  error_message_context
  max_errors
  dialect
  reset
  parse
  parse_into
  check_errors
  raise_error
  expression
  validate_expression
  errors
  sql

class ClickHouse.Generator(sqlglot.generator.Generator): View Source

    326    class Generator(generator.Generator):
    327        QUERY_HINTS = False
    328        STRUCT_DELIMITER = ("(", ")")
    329        NVL2_SUPPORTED = False
    330        TABLESAMPLE_REQUIRES_PARENS = False
    331
    332        STRING_TYPE_MAPPING = {
    333            exp.DataType.Type.CHAR: "String",
    334            exp.DataType.Type.LONGBLOB: "String",
    335            exp.DataType.Type.LONGTEXT: "String",
    336            exp.DataType.Type.MEDIUMBLOB: "String",
    337            exp.DataType.Type.MEDIUMTEXT: "String",
    338            exp.DataType.Type.TINYBLOB: "String",
    339            exp.DataType.Type.TINYTEXT: "String",
    340            exp.DataType.Type.TEXT: "String",
    341            exp.DataType.Type.VARBINARY: "String",
    342            exp.DataType.Type.VARCHAR: "String",
    343        }
    344
    345        TYPE_MAPPING = {
    346            **generator.Generator.TYPE_MAPPING,
    347            **STRING_TYPE_MAPPING,
    348            exp.DataType.Type.ARRAY: "Array",
    349            exp.DataType.Type.BIGINT: "Int64",
    350            exp.DataType.Type.DATETIME64: "DateTime64",
    351            exp.DataType.Type.DOUBLE: "Float64",
    352            exp.DataType.Type.ENUM: "Enum",
    353            exp.DataType.Type.ENUM8: "Enum8",
    354            exp.DataType.Type.ENUM16: "Enum16",
    355            exp.DataType.Type.FIXEDSTRING: "FixedString",
    356            exp.DataType.Type.FLOAT: "Float32",
    357            exp.DataType.Type.INT: "Int32",
    358            exp.DataType.Type.MEDIUMINT: "Int32",
    359            exp.DataType.Type.INT128: "Int128",
    360            exp.DataType.Type.INT256: "Int256",
    361            exp.DataType.Type.LOWCARDINALITY: "LowCardinality",
    362            exp.DataType.Type.MAP: "Map",
    363            exp.DataType.Type.NESTED: "Nested",
    364            exp.DataType.Type.NULLABLE: "Nullable",
    365            exp.DataType.Type.SMALLINT: "Int16",
    366            exp.DataType.Type.STRUCT: "Tuple",
    367            exp.DataType.Type.TINYINT: "Int8",
    368            exp.DataType.Type.UBIGINT: "UInt64",
    369            exp.DataType.Type.UINT: "UInt32",
    370            exp.DataType.Type.UINT128: "UInt128",
    371            exp.DataType.Type.UINT256: "UInt256",
    372            exp.DataType.Type.USMALLINT: "UInt16",
    373            exp.DataType.Type.UTINYINT: "UInt8",
    374        }
    375
    376        TRANSFORMS = {
    377            **generator.Generator.TRANSFORMS,
    378            exp.Select: transforms.preprocess([transforms.eliminate_qualify]),
    379            exp.AnyValue: rename_func("any"),
    380            exp.ApproxDistinct: rename_func("uniq"),
    381            exp.ArgMax: arg_max_or_min_no_count("argMax"),
    382            exp.ArgMin: arg_max_or_min_no_count("argMin"),
    383            exp.Array: inline_array_sql,
    384            exp.CastToStrType: rename_func("CAST"),
    385            exp.CurrentDate: lambda self, e: self.func("CURRENT_DATE"),
    386            exp.DateAdd: lambda self, e: self.func(
    387                "DATE_ADD", exp.Literal.string(e.text("unit") or "day"), e.expression, e.this
    388            ),
    389            exp.DateDiff: lambda self, e: self.func(
    390                "DATE_DIFF", exp.Literal.string(e.text("unit") or "day"), e.expression, e.this
    391            ),
    392            exp.Explode: rename_func("arrayJoin"),
    393            exp.Final: lambda self, e: f"{self.sql(e, 'this')} FINAL",
    394            exp.IsNan: rename_func("isNaN"),
    395            exp.Map: lambda self, e: _lower_func(var_map_sql(self, e)),
    396            exp.Nullif: rename_func("nullIf"),
    397            exp.PartitionedByProperty: lambda self, e: f"PARTITION BY {self.sql(e, 'this')}",
    398            exp.Pivot: no_pivot_sql,
    399            exp.Quantile: _quantile_sql,
    400            exp.RegexpLike: lambda self, e: f"match({self.format_args(e.this, e.expression)})",
    401            exp.StartsWith: rename_func("startsWith"),
    402            exp.StrPosition: lambda self, e: f"position({self.format_args(e.this, e.args.get('substr'), e.args.get('position'))})",
    403            exp.VarMap: lambda self, e: _lower_func(var_map_sql(self, e)),
    404            exp.Xor: lambda self, e: self.func("xor", e.this, e.expression, *e.expressions),
    405        }
    406
    407        PROPERTIES_LOCATION = {
    408            **generator.Generator.PROPERTIES_LOCATION,
    409            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    410            exp.PartitionedByProperty: exp.Properties.Location.POST_SCHEMA,
    411            exp.OnCluster: exp.Properties.Location.POST_NAME,
    412        }
    413
    414        JOIN_HINTS = False
    415        TABLE_HINTS = False
    416        EXPLICIT_UNION = True
    417        GROUPINGS_SEP = ""
    418
    419        # there's no list in docs, but it can be found in Clickhouse code
    420        # see `ClickHouse/src/Parsers/ParserCreate*.cpp`
    421        ON_CLUSTER_TARGETS = {
    422            "DATABASE",
    423            "TABLE",
    424            "VIEW",
    425            "DICTIONARY",
    426            "INDEX",
    427            "FUNCTION",
    428            "NAMED COLLECTION",
    429        }
    430
    431        def _any_to_has(
    432            self,
    433            expression: exp.EQ | exp.NEQ,
    434            default: t.Callable[[t.Any], str],
    435            prefix: str = "",
    436        ) -> str:
    437            if isinstance(expression.left, exp.Any):
    438                arr = expression.left
    439                this = expression.right
    440            elif isinstance(expression.right, exp.Any):
    441                arr = expression.right
    442                this = expression.left
    443            else:
    444                return default(expression)
    445            return prefix + self.func("has", arr.this.unnest(), this)
    446
    447        def eq_sql(self, expression: exp.EQ) -> str:
    448            return self._any_to_has(expression, super().eq_sql)
    449
    450        def neq_sql(self, expression: exp.NEQ) -> str:
    451            return self._any_to_has(expression, super().neq_sql, "NOT ")
    452
    453        def regexpilike_sql(self, expression: exp.RegexpILike) -> str:
    454            # Manually add a flag to make the search case-insensitive
    455            regex = self.func("CONCAT", "'(?i)'", expression.expression)
    456            return f"match({self.format_args(expression.this, regex)})"
    457
    458        def datatype_sql(self, expression: exp.DataType) -> str:
    459            # String is the standard ClickHouse type, every other variant is just an alias.
    460            # Additionally, any supplied length parameter will be ignored.
    461            #
    462            # https://clickhouse.com/docs/en/sql-reference/data-types/string
    463            if expression.this in self.STRING_TYPE_MAPPING:
    464                return "String"
    465
    466            return super().datatype_sql(expression)
    467
    468        def cte_sql(self, expression: exp.CTE) -> str:
    469            if expression.args.get("scalar"):
    470                this = self.sql(expression, "this")
    471                alias = self.sql(expression, "alias")
    472                return f"{this} AS {alias}"
    473
    474            return super().cte_sql(expression)
    475
    476        def after_limit_modifiers(self, expression: exp.Expression) -> t.List[str]:
    477            return super().after_limit_modifiers(expression) + [
    478                self.seg("SETTINGS ") + self.expressions(expression, key="settings", flat=True)
    479                if expression.args.get("settings")
    480                else "",
    481                self.seg("FORMAT ") + self.sql(expression, "format")
    482                if expression.args.get("format")
    483                else "",
    484            ]
    485
    486        def parameterizedagg_sql(self, expression: exp.Anonymous) -> str:
    487            params = self.expressions(expression, key="params", flat=True)
    488            return self.func(expression.name, *expression.expressions) + f"({params})"
    489
    490        def placeholder_sql(self, expression: exp.Placeholder) -> str:
    491            return f"{{{expression.name}: {self.sql(expression, 'kind')}}}"
    492
    493        def oncluster_sql(self, expression: exp.OnCluster) -> str:
    494            return f"ON CLUSTER {self.sql(expression, 'this')}"
    495
    496        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
    497            kind = self.sql(expression, "kind").upper()
    498            if kind in self.ON_CLUSTER_TARGETS and locations.get(exp.Properties.Location.POST_NAME):
    499                this_name = self.sql(expression.this, "this")
    500                this_properties = " ".join(
    501                    [self.sql(prop) for prop in locations[exp.Properties.Location.POST_NAME]]
    502                )
    503                this_schema = self.schema_columns_sql(expression.this)
    504                return f"{this_name}{self.sep()}{this_properties}{self.sep()}{this_schema}"
    505
    506            return super().createable_sql(expression, locations)

Generator converts a given syntax tree to the corresponding SQL string.

###### Arguments:

- pretty: Whether or not to format the produced SQL string. Default: False.
- identify: Determines when an identifier should be quoted. Possible values are: False (default): Never quote, except in cases where it's mandatory by the dialect. True or 'always': Always quote. 'safe': Only quote identifiers that are case insensitive.
- normalize: Whether or not to normalize identifiers to lowercase. Default: False.
- pad: Determines the pad size in a formatted string. Default: 2.
- indent: Determines the indentation size in a formatted string. Default: 2.
- normalize_functions: Whether or not to normalize all function names. Possible values are: "upper" or True (default): Convert names to uppercase. "lower": Convert names to lowercase. False: Disables function name normalization.
- unsupported_level: Determines the generator's behavior when it encounters unsupported expressions. Default ErrorLevel.WARN.
- max_unsupported: Maximum number of unsupported messages to include in a raised UnsupportedError. This is only relevant if unsupported_level is ErrorLevel.RAISE. Default: 3
- leading_comma: Determines whether or not the comma is leading or trailing in select expressions. This is only relevant when generating in pretty mode. Default: False
- max_text_width: The max number of characters in a segment before creating new lines in pretty mode. The default is on the smaller end because the length only represents a segment and not the true line length. Default: 80
- comments: Whether or not to preserve comments in the output SQL code. Default: True

QUERY_HINTS = False
STRUCT_DELIMITER = ('(', ')')
NVL2_SUPPORTED = False
TABLESAMPLE_REQUIRES_PARENS = False
STRING_TYPE_MAPPING = {<Type.CHAR: 'CHAR'>: 'String', <Type.LONGBLOB: 'LONGBLOB'>: 'String', <Type.LONGTEXT: 'LONGTEXT'>: 'String', <Type.MEDIUMBLOB: 'MEDIUMBLOB'>: 'String', <Type.MEDIUMTEXT: 'MEDIUMTEXT'>: 'String', <Type.TINYBLOB: 'TINYBLOB'>: 'String', <Type.TINYTEXT: 'TINYTEXT'>: 'String', <Type.TEXT: 'TEXT'>: 'String', <Type.VARBINARY: 'VARBINARY'>: 'String', <Type.VARCHAR: 'VARCHAR'>: 'String'}
TYPE_MAPPING = {<Type.NCHAR: 'NCHAR'>: 'CHAR', <Type.NVARCHAR: 'NVARCHAR'>: 'VARCHAR', <Type.MEDIUMTEXT: 'MEDIUMTEXT'>: 'String', <Type.LONGTEXT: 'LONGTEXT'>: 'String', <Type.TINYTEXT: 'TINYTEXT'>: 'String', <Type.MEDIUMBLOB: 'MEDIUMBLOB'>: 'String', <Type.LONGBLOB: 'LONGBLOB'>: 'String', <Type.TINYBLOB: 'TINYBLOB'>: 'String', <Type.INET: 'INET'>: 'INET', <Type.CHAR: 'CHAR'>: 'String', <Type.TEXT: 'TEXT'>: 'String', <Type.VARBINARY: 'VARBINARY'>: 'String', <Type.VARCHAR: 'VARCHAR'>: 'String', <Type.ARRAY: 'ARRAY'>: 'Array', <Type.BIGINT: 'BIGINT'>: 'Int64', <Type.DATETIME64: 'DATETIME64'>: 'DateTime64', <Type.DOUBLE: 'DOUBLE'>: 'Float64', <Type.ENUM: 'ENUM'>: 'Enum', <Type.ENUM8: 'ENUM8'>: 'Enum8', <Type.ENUM16: 'ENUM16'>: 'Enum16', <Type.FIXEDSTRING: 'FIXEDSTRING'>: 'FixedString', <Type.FLOAT: 'FLOAT'>: 'Float32', <Type.INT: 'INT'>: 'Int32', <Type.MEDIUMINT: 'MEDIUMINT'>: 'Int32', <Type.INT128: 'INT128'>: 'Int128', <Type.INT256: 'INT256'>: 'Int256', <Type.LOWCARDINALITY: 'LOWCARDINALITY'>: 'LowCardinality', <Type.MAP: 'MAP'>: 'Map', <Type.NESTED: 'NESTED'>: 'Nested', <Type.NULLABLE: 'NULLABLE'>: 'Nullable', <Type.SMALLINT: 'SMALLINT'>: 'Int16', <Type.STRUCT: 'STRUCT'>: 'Tuple', <Type.TINYINT: 'TINYINT'>: 'Int8', <Type.UBIGINT: 'UBIGINT'>: 'UInt64', <Type.UINT: 'UINT'>: 'UInt32', <Type.UINT128: 'UINT128'>: 'UInt128', <Type.UINT256: 'UINT256'>: 'UInt256', <Type.USMALLINT: 'USMALLINT'>: 'UInt16', <Type.UTINYINT: 'UTINYINT'>: 'UInt8'}
TRANSFORMS = {<class 'sqlglot.expressions.DateAdd'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.CaseSpecificColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CheckColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CollateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CommentColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DateFormatColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DefaultColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.EncodeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExternalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.HeapProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InlineLengthColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IntervalSpan'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LanguageProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LocationProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LogProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.MaterializedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NonClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NotForReplicationColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnCommitProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnUpdateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OutputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.PathColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ReturnsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SampleProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SettingsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.StabilityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TemporaryProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ToTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransientProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransformModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TitleColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UppercaseColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VarMap'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.VolatileProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.Select'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.AnyValue'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ApproxDistinct'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArgMax'>: <function arg_max_or_min_no_count.<locals>._arg_max_or_min_sql>, <class 'sqlglot.expressions.ArgMin'>: <function arg_max_or_min_no_count.<locals>._arg_max_or_min_sql>, <class 'sqlglot.expressions.Array'>: <function inline_array_sql>, <class 'sqlglot.expressions.CastToStrType'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.CurrentDate'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.DateDiff'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.Explode'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Final'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.IsNan'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Map'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.Nullif'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.PartitionedByProperty'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.Pivot'>: <function no_pivot_sql>, <class 'sqlglot.expressions.Quantile'>: <function _quantile_sql>, <class 'sqlglot.expressions.RegexpLike'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.StartsWith'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.StrPosition'>: <function ClickHouse.Generator.<lambda>>, <class 'sqlglot.expressions.Xor'>: <function ClickHouse.Generator.<lambda>>}
PROPERTIES_LOCATION = {<class 'sqlglot.expressions.AlgorithmProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.AutoIncrementProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BlockCompressionProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CharacterSetProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ChecksumProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CollateProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Cluster'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ClusteredByProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DataBlocksizeProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.DefinerProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.DictRange'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DictProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistStyleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.EngineProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExternalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.FallbackProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.FileFormatProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.FreespaceProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.HeapProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.InputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IsolatedLoadingProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.JournalProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.LanguageProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LikeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LocationProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LockingProperty'>: <Location.POST_ALIAS: 'POST_ALIAS'>, <class 'sqlglot.expressions.LogProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.MaterializedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.MergeBlockRatioProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.OnProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OnCommitProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.Order'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OutputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PartitionedByProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PartitionedOfProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PrimaryKey'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Property'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ReturnsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatDelimitedProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatSerdeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SampleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SchemaCommentProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SerdeProperties'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Set'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SettingsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SetProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.SortKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.StabilityProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TemporaryProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ToTableProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TransientProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.TransformModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.MergeTreeTTL'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.VolatileProperty'>: <Location.UNSUPPORTED: 'UNSUPPORTED'>, <class 'sqlglot.expressions.WithDataProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.WithSystemVersioningProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OnCluster'>: <Location.POST_NAME: 'POST_NAME'>}
JOIN_HINTS = False
TABLE_HINTS = False
EXPLICIT_UNION = True
GROUPINGS_SEP = ''
ON_CLUSTER_TARGETS = {'FUNCTION', 'NAMED COLLECTION', 'TABLE', 'INDEX', 'VIEW', 'DATABASE', 'DICTIONARY'}
def eq_sql(self, expression: sqlglot.expressions.EQ) -> str: View Source

    447        def eq_sql(self, expression: exp.EQ) -> str:
    448            return self._any_to_has(expression, super().eq_sql)

def neq_sql(self, expression: sqlglot.expressions.NEQ) -> str: View Source

    450        def neq_sql(self, expression: exp.NEQ) -> str:
    451            return self._any_to_has(expression, super().neq_sql, "NOT ")

def regexpilike_sql(self, expression: sqlglot.expressions.RegexpILike) -> str: View Source

    453        def regexpilike_sql(self, expression: exp.RegexpILike) -> str:
    454            # Manually add a flag to make the search case-insensitive
    455            regex = self.func("CONCAT", "'(?i)'", expression.expression)
    456            return f"match({self.format_args(expression.this, regex)})"

def datatype_sql(self, expression: sqlglot.expressions.DataType) -> str: View Source

    458        def datatype_sql(self, expression: exp.DataType) -> str:
    459            # String is the standard ClickHouse type, every other variant is just an alias.
    460            # Additionally, any supplied length parameter will be ignored.
    461            #
    462            # https://clickhouse.com/docs/en/sql-reference/data-types/string
    463            if expression.this in self.STRING_TYPE_MAPPING:
    464                return "String"
    465
    466            return super().datatype_sql(expression)

def cte_sql(self, expression: sqlglot.expressions.CTE) -> str: View Source

    468        def cte_sql(self, expression: exp.CTE) -> str:
    469            if expression.args.get("scalar"):
    470                this = self.sql(expression, "this")
    471                alias = self.sql(expression, "alias")
    472                return f"{this} AS {alias}"
    473
    474            return super().cte_sql(expression)

def after_limit_modifiers(self, expression: sqlglot.expressions.Expression) -> List[str]: View Source

    476        def after_limit_modifiers(self, expression: exp.Expression) -> t.List[str]:
    477            return super().after_limit_modifiers(expression) + [
    478                self.seg("SETTINGS ") + self.expressions(expression, key="settings", flat=True)
    479                if expression.args.get("settings")
    480                else "",
    481                self.seg("FORMAT ") + self.sql(expression, "format")
    482                if expression.args.get("format")
    483                else "",
    484            ]

def parameterizedagg_sql(self, expression: sqlglot.expressions.Anonymous) -> str: View Source

    486        def parameterizedagg_sql(self, expression: exp.Anonymous) -> str:
    487            params = self.expressions(expression, key="params", flat=True)
    488            return self.func(expression.name, *expression.expressions) + f"({params})"

def placeholder_sql(self, expression: sqlglot.expressions.Placeholder) -> str: View Source

    490        def placeholder_sql(self, expression: exp.Placeholder) -> str:
    491            return f"{{{expression.name}: {self.sql(expression, 'kind')}}}"

def oncluster_sql(self, expression: sqlglot.expressions.OnCluster) -> str: View Source

    493        def oncluster_sql(self, expression: exp.OnCluster) -> str:
    494            return f"ON CLUSTER {self.sql(expression, 'this')}"

def createable_sql( self, expression: sqlglot.expressions.Create, locations: DefaultDict) -> str: View Source

    496        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
    497            kind = self.sql(expression, "kind").upper()
    498            if kind in self.ON_CLUSTER_TARGETS and locations.get(exp.Properties.Location.POST_NAME):
    499                this_name = self.sql(expression.this, "this")
    500                this_properties = " ".join(
    501                    [self.sql(prop) for prop in locations[exp.Properties.Location.POST_NAME]]
    502                )
    503                this_schema = self.schema_columns_sql(expression.this)
    504                return f"{this_name}{self.sep()}{this_properties}{self.sep()}{this_schema}"
    505
    506            return super().createable_sql(expression, locations)

SELECT_KINDS: Tuple[str, ...] = ()

##### Inherited Members

sqlglot.generator.Generator
  Generator
  NULL_ORDERING_SUPPORTED
  LOCKING_READS_SUPPORTED
  WRAP_DERIVED_VALUES
  CREATE_FUNCTION_RETURN_AS
  MATCHED_BY_SOURCE
  SINGLE_STRING_INTERVAL
  INTERVAL_ALLOWS_PLURAL_FORM
  TABLESAMPLE_WITH_METHOD
  TABLESAMPLE_SIZE_IS_PERCENT
  LIMIT_FETCH
  LIMIT_ONLY_LITERALS
  RENAME_TABLE_WITH_DB
  INDEX_ON
  QUERY_HINT_SEP
  IS_BOOL_ALLOWED
  DUPLICATE_KEY_UPDATE_WITH_SET
  LIMIT_IS_TOP
  RETURNING_END
  COLUMN_JOIN_MARKS_SUPPORTED
  EXTRACT_ALLOWS_QUOTES
  TZ_TO_WITH_TIME_ZONE
  VALUES_AS_TABLE
  ALTER_TABLE_INCLUDE_COLUMN_KEYWORD
  UNNEST_WITH_ORDINALITY
  AGGREGATE_FILTER_SUPPORTED
  SEMI_ANTI_JOIN_WITH_SIDE
  COMPUTED_COLUMN_WITH_TYPE
  SUPPORTS_TABLE_COPY
  COLLATE_IS_FUNC
  DATA_TYPE_SPECIFIERS_ALLOWED
  ENSURE_BOOLS
  CTE_RECURSIVE_KEYWORD_REQUIRED
  SUPPORTS_SINGLE_ARG_CONCAT
  STAR_MAPPING
  TIME_PART_SINGULARS
  TOKEN_MAPPING
  PARAMETER_TOKEN
  RESERVED_KEYWORDS
  WITH_SEPARATED_COMMENTS
  EXCLUDE_COMMENTS
  UNWRAPPED_INTERVAL_VALUES
  EXPRESSIONS_WITHOUT_NESTED_CTES
  KEY_VALUE_DEFINITONS
  SENTINEL_LINE_BREAK
  pretty
  identify
  normalize
  pad
  unsupported_level
  max_unsupported
  leading_comma
  max_text_width
  comments
  dialect
  normalize_functions
  unsupported_messages
  generate
  preprocess
  unsupported
  sep
  seg
  pad_comment
  maybe_comment
  wrap
  no_identify
  normalize_func
  indent
  sql
  uncache_sql
  cache_sql
  characterset_sql
  column_sql
  columnposition_sql
  columndef_sql
  columnconstraint_sql
  computedcolumnconstraint_sql
  autoincrementcolumnconstraint_sql
  compresscolumnconstraint_sql
  generatedasidentitycolumnconstraint_sql
  generatedasrowcolumnconstraint_sql
  periodforsystemtimeconstraint_sql
  notnullcolumnconstraint_sql
  transformcolumnconstraint_sql
  primarykeycolumnconstraint_sql
  uniquecolumnconstraint_sql
  create_sql
  clone_sql
  describe_sql
  prepend_ctes
  with_sql
  tablealias_sql
  bitstring_sql
  hexstring_sql
  bytestring_sql
  rawstring_sql
  datatypeparam_sql
  directory_sql
  delete_sql
  drop_sql
  except_sql
  except_op
  fetch_sql
  filter_sql
  hint_sql
  index_sql
  identifier_sql
  inputoutputformat_sql
  national_sql
  partition_sql
  properties_sql
  root_properties
  properties
  with_properties
  locate_properties
  property_name
  property_sql
  likeproperty_sql
  fallbackproperty_sql
  journalproperty_sql
  freespaceproperty_sql
  checksumproperty_sql
  mergeblockratioproperty_sql
  datablocksizeproperty_sql
  blockcompressionproperty_sql
  isolatedloadingproperty_sql
  partitionboundspec_sql
  partitionedofproperty_sql
  lockingproperty_sql
  withdataproperty_sql
  withsystemversioningproperty_sql
  insert_sql
  intersect_sql
  intersect_op
  introducer_sql
  kill_sql
  pseudotype_sql
  objectidentifier_sql
  onconflict_sql
  returning_sql
  rowformatdelimitedproperty_sql
  withtablehint_sql
  indextablehint_sql
  historicaldata_sql
  table_sql
  tablesample_sql
  pivot_sql
  version_sql
  tuple_sql
  update_sql
  values_sql
  var_sql
  into_sql
  from_sql
  group_sql
  having_sql
  connect_sql
  prior_sql
  join_sql
  lambda_sql
  lateral_sql
  limit_sql
  offset_sql
  setitem_sql
  set_sql
  pragma_sql
  lock_sql
  literal_sql
  escape_str
  loaddata_sql
  null_sql
  boolean_sql
  order_sql
  cluster_sql
  distribute_sql
  sort_sql
  ordered_sql
  matchrecognize_sql
  query_modifiers
  offset_limit_modifiers
  after_having_modifiers
  select_sql
  schema_sql
  schema_columns_sql
  star_sql
  parameter_sql
  sessionparameter_sql
  subquery_sql
  qualify_sql
  union_sql
  union_op
  unnest_sql
  where_sql
  window_sql
  partition_by_sql
  windowspec_sql
  withingroup_sql
  between_sql
  bracket_sql
  all_sql
  any_sql
  exists_sql
  case_sql
  constraint_sql
  nextvaluefor_sql
  extract_sql
  trim_sql
  convert_concat_args
  concat_sql
  concatws_sql
  check_sql
  foreignkey_sql
  primarykey_sql
  if_sql
  matchagainst_sql
  jsonkeyvalue_sql
  formatjson_sql
  jsonobject_sql
  jsonarray_sql
  jsonarrayagg_sql
  jsoncolumndef_sql
  jsonschema_sql
  jsontable_sql
  openjsoncolumndef_sql
  openjson_sql
  in_sql
  in_unnest_op
  interval_sql
  return_sql
  reference_sql
  anonymous_sql
  paren_sql
  neg_sql
  not_sql
  alias_sql
  aliases_sql
  attimezone_sql
  add_sql
  and_sql
  xor_sql
  connector_sql
  bitwiseand_sql
  bitwiseleftshift_sql
  bitwisenot_sql
  bitwiseor_sql
  bitwiserightshift_sql
  bitwisexor_sql
  cast_sql
  currentdate_sql
  collate_sql
  command_sql
  comment_sql
  mergetreettlaction_sql
  mergetreettl_sql
  transaction_sql
  commit_sql
  rollback_sql
  altercolumn_sql
  renametable_sql
  altertable_sql
  add_column_sql
  droppartition_sql
  addconstraint_sql
  distinct_sql
  ignorenulls_sql
  respectnulls_sql
  intdiv_sql
  dpipe_sql
  div_sql
  overlaps_sql
  distance_sql
  dot_sql
  propertyeq_sql
  escape_sql
  glob_sql
  gt_sql
  gte_sql
  ilike_sql
  ilikeany_sql
  is_sql
  like_sql
  likeany_sql
  similarto_sql
  lt_sql
  lte_sql
  mod_sql
  mul_sql
  nullsafeeq_sql
  nullsafeneq_sql
  or_sql
  slice_sql
  sub_sql
  trycast_sql
  log_sql
  use_sql
  binary
  function_fallback_sql
  func
  format_args
  text_width
  format_time
  expressions
  op_expressions
  naked_property
  set_operation
  tag_sql
  token_sql
  userdefinedfunction_sql
  joinhint_sql
  kwarg_sql
  when_sql
  merge_sql
  tochar_sql
  dictproperty_sql
  dictrange_sql
  dictsubproperty_sql
  clusteredbyproperty_sql
  anyvalue_sql
  querytransform_sql
  indexconstraintoption_sql
  indexcolumnconstraint_sql
  nvl2_sql
  comprehension_sql
  columnprefix_sql
  opclass_sql
  predict_sql
  forin_sql
  refresh_sql
  operator_sql